import java.io.*;
import java.lang.Math;
import java.util.*;

// -----------------============== NOTE =============--------------------
// To make changes to "terminal" data (for grammars other than XQuery),
// go to the "Generator.java" file--that is the ONLY file that needs to
// be modified for those purposes; everything else is designed to run
// independently of the grammar being used.
// -----------------=================================--------------------

class DefListNode extends GrammarNode {

    // -------------------------------------------------------
    // 		-=-=-=- DefListNode -=-=-=-
    // -------------------------------------------------------
    // A rule for a given grammar consists of a head and tail,
    // separated by a "::=" symbol; the 'DefListNode' of a
    // a rule corresponds to the tail.  That is, it is a LIST of
    // of the different syntaxes that DEFINE the head.  Since
    // each syntax is in turn a list of tokens, the DefListNode
    // ends up being a list of TokenListNodes.  Different syntaxes
    // (TokenListNodes) are separated by the vertical bar ("|")
    // in a proper grammar.  Thus, in the following rule,
    //
    // fake ::= "def1" | "{" def2 "}" | "def3" def4+
    //
    // The head is "fake", while the DefListNode is a list (vector)
    // containing three TokenListNodes:
    //
    //	myTokenLists[1] = "def1"
    //	myTokenLists[2] = "{" def2 "}"
    //	myTokenLists[3] = "def3" def4+
    //
    // -------------------------------------------------------

    // ************--==== Class Variables ====--************

    // Vector of TokenListNodes.
    private Vector myTokenLists;

    // Number of syntax choices (i.e. # of TokenListNodes).
    private int syntaxCnt;

    // TOTAL number of occurence indicators that appear in
    // myTokenLists.
    private int recurCount;

    // Syntax sums generated from the config file for purposes
    // of selecting one of the available syntax choices for
    // generation.
    private int [] partialSynSum;

    // **********--==== End Class Variables ====--**********

    // ----

    /* ==========================================================
     *
     * Constructor.
     *
     * -------- */

    public DefListNode(Vector V) {

        myTokenLists = V;
	syntaxCnt = 1;
	recurCount = 0;

    } // end constructor.

    // ----

    /* ==========================================================
     *
     * method unparse():
     *
     * Build and process this DefListNode structure based on inputs
     * and gathered stats, and then echo the structure to file.
     *
     * -------- */
 
    public void unparse(PrintWriter p, int indent) {

	boolean first = true;
	Iterator itr = myTokenLists.iterator();
	while (itr.hasNext()) {
	    if (!first) {
		p.print("\n\t| ");
		syntaxCnt++;
	    } // end if.
	    TokenListNode tmpNode = (TokenListNode)itr.next();
	    tmpNode.unparse(p,0);
	    recurCount += tmpNode.recurs();
	    first = false;
        } // end while.

	// Check for NESTED definition lists (i.e. those that occur
	// withIN a given syntax choice), and "unfold" (expand
	// out) those definitions into proper (flat) syntaxes.
	// Then, make this new, unfolded list be this node's actual
	// def list.

	Vector nested = unfold();
	if (nested.size() > 0) {
	// In the event that we replaced something, replace old list
	// with new list.
	    myTokenLists.clear();
	    myTokenLists.addAll(nested);
	    syntaxCnt = myTokenLists.size();
	} // end if.

	// we're done.
	return;

    } // end method unparse.

    // ----

    /* ==========================================================
     *
     * method unfold():
     *
     * "Unfolds" a DefListNode by looking at all TokenListNodes
     * and, for each one, finding any nested DefListNodes (i.e.
     * lists of tokens separated by vertical bars) and expanding
     * them out into multiple TokenListNodes so that each TokenListNode
     * has only ONE syntax choice contained within it.
     *
     * Ex.   The single syntax choice (TokenListNode)
     *		"def1" (def2 | def3) "def4"
     * 	     when put in UNFOLDED form, becomes TWO syntax choices:
     *		"def1" (def2) "def4",
     *		"def1" (def3) "def4".
     *
     *   Thus, 'myTokenLists' goes from being size 1 to being size 2,
     *   and ALL of the resultant TokenListNodes contain only one
     *   possible syntax.  Technically, '(def2)' and '(def3)' are
     *   still DefListNodes, but each only contains ONE possible
     *   syntax choice (as evident by the lack of vertical bars).
     *
     * -------- */

    public Vector unfold() {

	Vector result = new Vector();
	Iterator itr = myTokenLists.iterator();
	while (itr.hasNext()) {
	   result.addAll(((TokenListNode)itr.next()).unfold());
	} // end while.

	return result;

    } // end method unfold.

    // ----

    /* ==========================================================
     *
     * method distribute():
     *
     * Takes an occurence indicator and distributes it throughout
     * this DefListNode's TokenListNodes.  More specifically,
     * it turns each TokenListNode into its OWN DefListNode, and
     * then creates a TokenNode that consists of the new DefListNode
     * plus a TokenOccurNode built from the received occurence
     * indicator.  This is used by the TokenListNode's "unfold"
     * method to correctly assign occurence indicators during the
     * unfolding process.
     *
     * Ex.  fake ::= (def2 | def3)+
     *   When distributed, this becomes:
     *	    fake ::= ( (def2)+ | (def3)+ )
     *   After unfolding, this then becomes:
     *	    fake ::= (def2)+
     * 		  |  (def3)+
     *
     * -------- */

    public void distribute(char occur) {

	if (occur == NO_RECURRENCE) {
	// don't do anything.
	    return;
	} // end if.

	Vector tmpVec;		// TokenListNodes.
	Vector tmpVec2;		// TokenNodes.

	Vector tmpTokList = new Vector();
	DefListNode tmpDefList;
	TokenContentNode tmpContent;
	TokenOccurNode tmpOccur;
	TokenNode tmpToken;

	Iterator itr = myTokenLists.iterator();
	while (itr.hasNext()) {

	    // Create new storage vectors.
	    tmpVec = new Vector();
	    tmpVec2 = new Vector();

	    // Make this TokenListNode into a DefListNode.
	    tmpVec.add((TokenListNode)itr.next());
	    tmpDefList = new DefListNode(tmpVec);

	    // Make a TokenNode from the DefListNode and the
	    // received occurrence indicator.
	    tmpContent = new TokenContentNode(tmpDefList);
	    tmpOccur = new TokenOccurNode(occur);
	    tmpToken = new TokenNode(tmpContent,tmpOccur);

	    // Make the new TokenNode into a TokenListNode,
	    // and store it.
	    tmpVec2.add(tmpToken);
	    tmpTokList.add(new TokenListNode(tmpVec2));

	} // end while.

	// Replace old token list with 'distributed' token list.
	myTokenLists.clear();
	myTokenLists.addAll(tmpTokList);

	return;

    } // end method distribute.

    // ----

    /* ==========================================================
     *
     * method createConfigFile():
     *
     * Creates a configuration file for the grammar in question,
     * plugging default values into all the correct places and
     * leaving it up to the user to modify the configuration file
     * as needed for specific kind of language generations.
     *
     * -------- */

    public void createConfigFile(PrintWriter p, String head) {

	if ((recurs() == 0) && (syntaxCnt < 2)) {
	// then we don't need to add anything to the config file.
	    return;
	} // end if.

	// Else...

	Iterator itr = myTokenLists.iterator();
	TokenListNode tmpNode;
	String occurSymbols;
	String str;
	int cnt = 1;
	while (itr.hasNext()) {

	    tmpNode = (TokenListNode)itr.next();
	    occurSymbols = tmpNode.recurSymbols();
	    str = (head.toUpperCase() + "_SYNTAX_" + cnt);
	    p.print(str + prettySpaces(str) + "1");
	    p.println("  \t//  " + tmpNode.toString());
	    for (int i = 0; i < tmpNode.recurs(); i++) {
		str = (head.toUpperCase() + "_RECUR_" + cnt + "_" + (i+1));
		p.println(str + prettySpaces(str) +
			DEFAULT_RECUR_PROB + "  \t//  --");
		if (occurSymbols.charAt(i) != '?') {
		// then we need a parameter for the 'max'.
		    str += "_MAX";
		    p.println(str + prettySpaces(str) +
			DEFAULT_SYNTAX_WEIGHT + "  \t//  --");
		} // end inner if.
	    } // end for.
	    cnt++;

	} // end while.

	if (syntaxCnt > 1) {
	// then we need to specify a termination syntax for when we've hit
	// the max depth and want to quit.
	    str = (head.toUpperCase() + "_ENDING");
	    p.println(str + prettySpaces(str) +
		DEFAULT_TERM_SYNTAX + "  \t// ===> Termination Syntax.");
	} // end if.

	p.println();
	return;

    } // end method createConfigFile.

    // ----

    /* ==========================================================
     *
     * method sumUpSyntaxes():
     *
     * Goes through and creates an array of sums built according to
     * the syntax weights specified in the config file.  For the i-th
     * element in the array, the value stored is the sum of ALL
     * PRECEDING syntax weights, plus the i-th weight.  Thus, in
     * the following config example,
     *
     *		FAKE_SYNTAX_1	1	// def1
     *		FAKE_SYNTAX_2	2	// def2
     *		FAKE_SYNTAX_3	1	// def3
     *		FAKE_SYNTAX_4	3	// def4
     *
     * this method would return an array of 4 integers (one for
     * each syntax choice).  The first value (syntaxSums[0]) would
     * be 1, the second would be 3 (1+2), the third would be 4
     * (1+2+1), etc.  This array is used later to choose a random
     * syntax for generation--syntaxes with larger weights contribute
     * more to the final sum, and so are more likely to be generated.
     * For example, the total sum of the above syntaxes is 7; if we
     * then pick a random number between 1 and 7, we can determine
     * which syntax corresponds to that number: 1 => def1, [2,3] =>
     * def2, 4 => def3, and [5,6,7] => def4.  Thus, def2 is twice
     * as likely to be chosen as def1 and def3; def4 is three times
     * as likely as the same.
     * 
     * -------- */

    public int [] sumUpSyntaxes(String head, Vector allowed) {

	if (allowed.size() == 0) {
	    System.out.println("-=- ERROR: Found zero syntaxes "
		+ "allowed for " + head + "! -=-\n");
	    return null;
	} // end if.

	int [] syntaxSums = new int[myTokenLists.size()];
	if (myTokenLists.size() < 2) {
	// then there's nothing to add up (no entries in the
	// config file, because we only need weights if we have
	// 2 or more syntax choices).
	    syntaxSums[0] = 0;
	    return syntaxSums;
	} // end if.

	// Else...

	String str;
	int cnt = 1;
	int sum = 0;
	Double tmpDbl = null;
	Iterator itr = myTokenLists.iterator();
	while (itr.hasNext()) {

	    if (!allowed.contains(new Integer(cnt))) {
	    // then don't include this in the sum.
		itr.next();
		syntaxSums[cnt-1] = sum;
		cnt++;
		continue;
	    } // end if.

	    TokenListNode tmpNode = (TokenListNode)itr.next();
	    str = (head.toUpperCase() + "_SYNTAX_" + cnt);
	    tmpDbl = (Double)configMap.get(str);
	    if (tmpDbl == null) {
		System.out.println("-=- ERROR: Failed map lookup: " +
		 		str + " -=-\n");
	    } // end if.
	    else {
		sum += (int)((Double)configMap.get(str)).doubleValue();
		syntaxSums[cnt-1] = sum;
	    } // end else.
	    cnt++;
	} // end while.

	if (sum == 0) {
	// then we have more than one choice, but all are zero-weighted,
	// which means we don't know what to generate.  Return an indicator;
	// this will be caught later.
	    syntaxSums[cnt-2] = -1;
	} // end if.

	// Return what we have.
	return syntaxSums;

    } // end method sumUpSyntaxes.

    // ----

    /* ==========================================================
     *
     * method terminate():
     *
     * Returns the "termination syntax" for this DefListNode, which
     * is the syntax choice that should ALWAYS be chosen when we
     * have exceeded the maxDepth of our generation and want to quit
     * as soon as possible.  The particular syntax to choose is
     * set in (and read from) the configuration file, as is the
     * maxDepth parameter telling when to choose that syntax.  If
     * no termination syntax is provided, or if the one found is
     * invalid, then a '-1' is returned.
     *
     * -------- */

    public int terminate(String head) {

	Double dbl = (Double)configMap.get(head.toUpperCase() + "_ENDING");
	if (dbl == null) {
	// couldn't find the termination syntax parameter.
	    System.out.println("-=- ERROR: Failed map lookup: " +
		head.toUpperCase() + "_ENDING. -=-");
	    return -1;
	} // end if.

	// Else if termination syntax choice is invalid, ignore it.
	if (dbl.doubleValue() > numChoices()) {
	    System.out.println("[ Ignoring invalid termination syntax" +
		"(" + dbl.doubleValue() + ") for " + head + ". ]");
	    return -1;
	} // end if.

	// Else, we're good; return the termination syntax.
	return (int)(dbl.doubleValue());

    } // end method terminate.

    // ----

    /* ==========================================================
     *
     * method generate():
     *
     * Takes a number corresponding to some syntax choice, loads
     * the appropriate TokenListNode, and then makes the call to
     * "generate" that TokenListNode, where by 'generate' we mean
     * create a result string that, while random, nonetheless
     * conforms to the syntax specified by the TokenListNode.
     *
     * -------- */

    public GResult generate(int syntax, int origSyntax,
		String head, int occurID, int nestedListCnt,
		Vector constraints) {

	TokenListNode tokList = getDefinition(syntax);

	if (tokList == null) {
	   System.out.println("-=- ERROR: Invalid Definition Number!\n");
	   GResult res = new GResult();
	   res.typeVal = "NONE";
	   res.strVal = "UNAVAILABLE";
	   return res;
	} // end if.

	if (syntax == -1) {
	// then sum of all syntax weights was zero; don't generate
	// anything.
	   GResult res = new GResult();
	   res.typeVal = "NONE";
	   res.strVal = "";
	   return res;
	} // end if.

	// Else, generate the specified syntax.
	return tokList.generate(head, syntax+1, origSyntax,
		occurID, nestedListCnt, constraints);

    } // end method generate.

    // ----

    /* ==========================================================
     *
     * method fullInsert():
     *
     * Takes a vector of TokenNodes and inserts those TokenNodes at
     * the front of ALL TokenListNodes in this DefListNode.  This
     * is used during the 'unfold' phase of parsing, and is called
     * by the "unfold" method declared inside the TokenListNode
     * objects.
     *
     * -------- */

    public void fullInsert(Vector nodes) {

	Iterator itr = myTokenLists.iterator();
	while (itr.hasNext()) {
	// insert nodes at FRONT.
	    ((TokenListNode)itr.next()).insert(nodes,FRONT);
	} // end while.

	return;

    } // end method fullInsert.

    // ----

    /* ==========================================================
     *
     * method getDefinition():
     *
     * Returns the TokenListNode (syntax choice) corresponding to the
     * received number, where counting begins at 0.  Thus, "0" refers
     * to the first syntax choice, "1" to the second, etc...
     *
     * -------- */

    public TokenListNode getDefinition(int num) {

	int cnt = 0;
	Iterator itr = myTokenLists.iterator();
	TokenListNode tmpNode;
	while (itr.hasNext()) {
	    tmpNode = (TokenListNode)itr.next();
	    if (cnt == num) {
		return tmpNode;
	    } // end if.
	    cnt++;
        } // end while.

	// If we get here, then the desired rule doesn't exist.
	return null;

    } // end method getTail.

    // ----

    /* ==========================================================
     *
     * method syntax():
     *
     * Returns the string version of the TokenListNode (syntax
     * choice) corresponding to the received number, where counting
     * begins at 0.  Thus, "0" refers to the first syntax choice,
     * "1" to the second, etc...This method differs from the
     * "getDefinition" method in that here we only return the String
     * version of the TokenListNode, whereas in "getDefinition",
     * we actually return the TokenListNode itself.
     *
     * -------- */

    public String syntax(int num) {

	int cnt = 0;
	TokenListNode tmpNode;
	Iterator itr = myTokenLists.iterator();
	while (itr.hasNext()) {
	    tmpNode = (TokenListNode)itr.next();
	    if (cnt == num) {
		return (tmpNode.toString());
	    } // end if.
	    cnt++;
        } // end while.

	// If we get here, then the desired rule doesn't exist.
	return "UNAVAILABLE";

    } // end method syntax.

    // ----

    /* ==========================================================
     *
     * method allChoices():
     *
     * Returns a vector of Integer objects, where each Integer
     * simply corresponds to one of the syntax choices.
     *
     * -------- */

    public Vector allChoices() {

	Vector result = new Vector();
	for (int i = 0; i < numChoices(); i++) {
	    result.add(new Integer(i+1));
	} // end for.

	return result;

    } // end method allChoices.

    /* ==========================================================
     *
     * method toString():
     *
     * Returns a String representation of this DefListNode.
     *
     * -------- */

    public String toString() {

	StringBuffer str = new StringBuffer();
	Iterator itr = myTokenLists.iterator();
	boolean first = true;
	while (itr.hasNext()) {
	    if (!first) {
		str.append("| ");
	    } // end if.
	    str.append(((TokenListNode)itr.next()).toString());
	    first = false;
	} // end while.

	return str.toString();

    } // end toString.

    // ----

    /* ==========================================================
     *
     * method prettySpaces():
     *
     * Returns a String of spaces used for formatting the config
     * configuration file; the length of this string is determined 
     * by the length of the received string.
     *
     * -------- */

    public String prettySpaces(String someStr) {

	final int MAXLEN = 45;

	int tmp = MAXLEN - someStr.length();
	String str = "";
	for (int i = 0; i < tmp; i++)
	   str += " ";

	return str;

    } // end method prettySpaces.

    // ----

    /* ==========================================================
     *
     * [ Other Helper Functions ]:
     *
     * The remaining functions simply allow access to this
     * DefListNode's data and statistics.
     *
     * -------- */

    public Vector tokLists() {

	return myTokenLists;

    } // end method tokLists.

    // ----

    public int numChoices() {

	return myTokenLists.size();

    } // end method numChoices.

    // ----

    public int recurs() {

	return recurCount;

    } // end method recurs.

    // ----

    public String getRecurs() {

	String result = "";
	Iterator itr = myTokenLists.iterator();
	while (itr.hasNext()) {
	    result += ((TokenListNode)itr.next()).recurSymbols();
	} // end while.

	return result;

    } // end method getRecurs.

    // ----

    public int numNestedLists() {

	int cnt = 0;
	Iterator itr = myTokenLists.iterator();
	while (itr.hasNext()) {
	    cnt += ((TokenListNode)itr.next()).countNestedLists();
	} // end while.

	return cnt;

    } // end method numNestedLists.

} // end class DefListNode.
