import java.io.*;
import java.util.*;
import java.util.Random;

/* ***********************************************************************
 * =====================-------------------------------===================
 ## THIS IS THE ONLY FILE THAT NEEDS TO BE ALTERED FOR DIFFERENT
 ## GRAMMARS!! (everything else is designed to run independently of the
 # grammar being used).
 * =====================-------------------------------===================
 *
 * This file is used to access and generate data that is EXTERIOR to the
 * Grammar file.  More specifically, it allows the generation of terminals
 * that are NOT specifically defined within the Grammar itself.  These
 * terminals are IdNodes for which NO RULE in the grammar has a matching
 * head.
 *
 * For example, in the case of an XQuery grammar, we might include a
 * terminal called "eltName" in our grammar, but choose NOT to define it.
 * Then, when we try to generate such a name, we will fail to find a rule
 * and so will call the "checkTerminal" method with "eltName" as a
 * parameter.  That method will in turn access a "terminalData" Object
 * containing a list of all of the valid Element names, and will pick
 * one of them as our result (instead of just randomly generating some
 * element name that doesn't really exist.  To access the data in said
 * file, we just use methods defined in this "Generator" class.
 *
 * While most of the methods in this class can be replaced for different
 * grammars, there are a couple that MUST be included if the rest of the
 * program is to work correctly.  Those methods are defined first, and
 * are clearly marked.  The remaining methods may (and probably will)
 * need to be changed if use of a grammar requires different external
 * data.
 *
 * There are no restrictions on the format of the data in the terminal
 * file, nor on the structure of that data as it is read into memory--
 * we just modify the methods of this class to handle the different
 * formats.
 * 
 * =====================-------------------------------===================
 * *********************************************************************** */

public class Generator {

    // ************--==== Class Variables ====--************

    // Basic [ ### required ### ]
    protected int seed;
    protected Object terminalData;
    protected Random random;

    // Possible components of a rule.
    protected static final int DEF_LIST = 0;
    protected static final int CHARSET = 1;
    protected static final int STRING_LIT = 2;
    protected static final int ID = 3;

    // Other (non-required).
    protected int varCnt = 1;
    protected int funcCnt = 1;
    protected int eltCnt = 1;
    protected int attrCnt = 1;
    protected int genIdCnt = 1;
    protected int exprCnt = 0;

    Vector noSpaceBefore;
    Vector noSpaceAfter;

    protected Map funcMap;
    protected Stack eltNameStack;

    // Other constants.
    protected static final int MAX_CDATA_LEN = 15;
    protected static final int NO_TYPE = -2;
    protected static final int NEEDS_TYPE = -1;

    // **********--==== End Class Variables ====--**********

    // ----

 // ####################################################################
 // =================------- 	REQUIRED METHODS ------=================
 // ####################################################################
 //
 // ----
 // The following methods can be changed as needed, but they ** CANNOT **
 // be deleted, and all interfaces ** MUST ** be preserved (i.e. all
 // methods must accept the parameters used below, and must return the
 // specified type of value).
 // ----

    /* ==========================================================
     *
     * Constructor.
     * [ ### required ### ]
     *
     * -------- */

    public Generator(Object obj)  { 

	terminalData = obj;
	funcMap = new HashMap();
	eltNameStack = new Stack();

	// Load up some standard XQuery function names.
	loadFuncs();

	// Load "space" info.
	loadSpaces();

    } // end constructor.

    // ----

    /* ==========================================================
     *
     * method loadTerminalData():
     * [ ### required ### ]
     *
     * Must take a file, read data from that file into some Object,
     * and then return that object.  In the XQuery case, the Object
     * is a Vector, but this is NOT required: we can alter the way
     * the file is read/stored as much we like, as long as the
     * terminalData Object can be loaded from the received file and
     * is returned by this method.  Obviously, the other methods
     * in this class would have to be modified, as well, to handle
     * the new type of terminalData Object.
     *
     * -------- */

    public static Object loadTerminalData(FileReader inFile) {

	// Two vectors: one for elements, one for attributes.
	Vector [] result = new Vector[2];
	result[0] = new Vector();
	result[1] = new Vector();

	int type = -1;
	String line = readLine(inFile);
	while (!line.equals("EOF")) {

	    line = line.trim();
	    if (line.equals("-- Elements --")) {
	    // then we're loading elements.
		type = 0;
		line = readLine(inFile);
		continue;
	    } // end if.
	    else if (line.equals("-- Attributes --")) {
	    // we're loading attributes.
		type = 1;
		line = readLine(inFile);
		continue;
	    } // end else if.
	    else if (line.length() == 0) {
		line = readLine(inFile);
		continue;
	    } // end else if.

	    // Else, read and store the name.
	    result[type].add(line);
	    line = readLine(inFile);

	} // end while.

	return result;

    } // end method loadTerminalData.

    // ----

    /* ==========================================================
     *
     * method readLine():
     * [ ### required ### ]
     *
     * Reads a line from the received FileReader and returns it.
     *
     * -------- */

    static String readLine(FileReader aFile) {
 
	String str = "";
	try {
	    int tmpChar = aFile.read();
	    if (tmpChar == -1) {
		return "EOF";
	    } // end if.
	    while ((tmpChar != -1) && ((char)tmpChar != '\n')) {
		str += (char)tmpChar;
		tmpChar = aFile.read();
	    } // end while.
	} // end try.
	catch (IOException e) {
	    e.printStackTrace();
	} // end catch.

	return str;
 
    } // end method readLine.

    /* ==========================================================
     *
     * method checkTerminal():
     * [ ### required ### ]
     *
     * Takes the name of some 'terminal' head and returns a String
     * value corresponding to it.
     *
     * How exactly the resultant string value is obtained depends
     * on the type of the "terminalData" Object.  This method can
     * be altered as needed for different types, as long as it
     * continues to take a terminal Id (in the form of a String) and
     * return some String value.  Obviously, this method's treatment
     * of the terminalData Object will have to correspond with that
     * of the other methods (esp. the loadTerminalData() method).
     *
     * NOTE: In the case of a failed lookup, the value "NONE" (in
     * all caps) should be returned.
     *
     * -------- */

    public String checkTerminal(String head, Vector constraints,
	Vector subtypes) {

	Vector [] terms = (Vector[])(terminalData);
	if (terms == null) {
	// nothing to check; return 'NONE'.
	    return "NONE";
	} // end if.

	if (head.equals("FunctionName")) {
	    return (getExistingFuncName());
	} // end if
	else if (head.equals("Argument")) {
	    return (getGenericId());
	} // end else.
	else if (head.equals("ElementName")) {
	    int which = random.nextInt(terms[0].size());
	    return ((String)(terms[0].elementAt(which)));
	} // end else if.
	else if (head.equals("AttributeName")) {
	// Need to check/enforce our type constraints.
	    Vector choices = getValidAttributes(constraints,
		head, subtypes);
	    if (choices.size() < 1) {
	    // no valid choices; return indicator.
		return "INVALID";
	    } // end if.
	    int which = random.nextInt(choices.size());
	    return ((String)(choices.elementAt(which)));
	} // end else if.
	else if (head.equals("NCName")) {
	    return (getGenericId());
	} // end else.
	else if (head.equals("QName")) {
	    return (getGenericId());
	} // end else.
	else if (head.equals("Cdata")) {
	    return (getCDATA());
	} // end else.
	else if (head.equals("DoubleQuoteLiteral")) {
	    return ("\"" + getCDATA() + "\"");
	} // end else.
	else if (head.equals("SingleQuoteLiteral")) {
	    return ("'" + getCDATA() + "'");
	} // end else.
	else if (head.equals("UnquotedStringLiteral")) {
	    return (getCDATA());
	} // end else.
	else if (head.equals("Integer")) {
	    return ("" + (random.nextInt()));
	} // end else.
	else if (head.equals("Float")) {
	    return ("" + (random.nextFloat()));
	} // end else.
	else if (head.equals("Double")) {
	    return ("" + (random.nextDouble()));
	} // end else.
	else if (head.equals("Variable")) {
	    return (getNewVarName());
	} // end else.
	else if (head.equals("TagStartEltName")) {
	    int which = random.nextInt(terms[0].size());
	    String str = ((String)(terms[0].elementAt(which)));
	    eltNameStack.push(str);
	    return (str);
	} // end else.
	else if (head.equals("TagEndEltName")) {
	    return ((String)eltNameStack.pop());
	} // end else.

	return "NONE";

    } // end method checkTerminals.

    // ----

    /* ==========================================================
     *
     * method getTypeOfTerminal():
     * [ ### required ### ]
     *
     * For some received terminal string, finds and returns
     * the 'type' of that terminal as a string.
     *
     * -------- */
 
    public String getTypeOfTerminal(String term, String head) {

	if (head.equals("AttributeName")) {
	// type is inside "term"; pull it out.
	    int spaceLoc = term.indexOf(" ");
	    if (spaceLoc < 0) {
		System.out.println("-=- ERROR: Invalid attribute spec " +
		    "in terminal data file: " + term);
		return head;        // Default = head.
	    } // end inner if.
	    // Else, return the type.
	    return term.substring(spaceLoc + 1, term.length());
	} // end outer if.

	// Default: set type to be head of the terminal.
	return head;
 
    } // end method getTypeOfTerminal.

    // ----

    /* ==========================================================
     *
     * method getValueOfTerminal():
     * [ ### required ### ]
     *
     * For some received terminal string, finds and returns
     * the 'value' of that terminal as a string (i.e. whatever is
     * going to actually appear in the final, generated sentence).
     *
     * -------- */
 
    public String getValueOfTerminal(String term, String head) {

	if (head.equals("AttributeName")) {
	// value is inside "term"; pull it out.
	    int spaceLoc = term.indexOf(" ");
	    if (spaceLoc < 0) {
		return term;        // Default: let value = term.
	    } // end inner if.
	    // Else, return the type.
	    return term.substring(0, spaceLoc);
	} // end outer if.

	// Default: assume value is just 'term'.
	return term;

    } // end method getValueOfTerminal.

    // ----

    /* ==========================================================
     *
     * method getSpacingBetween():
     * [ ### required ### ]
     * 
     * Takes two nodes, returns the spacing required between them.
     * [ This needs to be changed to deal with spaces better ]
     *
     * -------- */

    public String getSpacingBetween(String prev, String curr) {

	if (prev == null) {
	// then 'curr' is the first token in this generation; no
	// spacing required before it.
	    return "";
	} // end if.

	// -=-=- For TERMINALS that are NOT defined in the grammar
	// -=-=- itself, check for the terminal ITSELF, like
	// -=-=- the following (quotes should be included in the check).

	if (prev.equals("\"<\"") && curr.equals("TagStartEltName")) {
	    // for tag open, no space after open brace ('<').
	    return "";
	} // end if.

	// -=-=- For all other terminals, check for the STRING LITERAL
	// -=-=- (with quotes) that will ultimately result.

	if (curr.equals("\">\"")) {
	// for tag close, no space after element name.
	    return "";
	} // end if.

	if (prev.equals("\"$\"")) {
	// for var names, no space after dollar sign.
	    return "";
	} // end if.

	if (noSpaceAfter.contains(prev) ||
	   noSpaceBefore.contains(curr)) {
	    return "";
	} // end if.

	if (isPunctuation(prev) || isPunctuation(curr)) {
	// no spacing required before/after most puncutation.
	    return "";
	} // end if.

	// Default is to return a space.
	return " ";

    } // end method getSpacingBetween.

    // ----

    /* ==========================================================
     *
     * method setRandom():
     * [ ### required ### ]
     *
     * Simply initializes the class "random" variable according to
     * whatever value is currently in the class "seed" variable.
     *
     * -------- */

    public void setRandom() { 

	this.random = new Random(seed);
	return;

    } // end method setRandom.

    // ----

    /* ==========================================================
     *
     * method getRandom():
     * [ ### required ### ]
     *
     * Returns the class random object.
     *
     * -------- */

    public Random getRandom() { 

	return random;

    } // end method setRandom.

    // ----

    /* ==========================================================
     *
     * method setSeed():
     * [ ### required ### ]
     *
     * Sets the class "seed" variable to the value received.  If
     * the received value is '0', then a seed is randomly
     * generated and stored.
     *
     * -------- */

    public int setSeed(int someSeed) {

	if (someSeed != 0) {
	    seed = someSeed;
	} // end if.
	else {
	// pick a seed at random; we have to do this
	// so that we can echo it to the screen, which
	// will allow the user to duplicate a particular
	// generation with ease.
	    seed = new Random().nextInt();
	} // end else.

	return seed;

    } // end method setSeed.

    // ----

    /* ==========================================================
     *
     * method resetCnts():
     * [ ### required ### ]
     *
     * Use this method to reset any counters/variables/etc. that
     * need to be initialized at the start of EVERY Query generation
     * (where by "Query" we just mean the VERY first rule in the
     * grammar file, from which everything else is built).
     *
     * -------- */

    public void resetCnts() {

	varCnt = 1;
	funcCnt = 1;
	eltCnt = 1;
	attrCnt = 1;
	genIdCnt = 1;
	exprCnt = 0;

    } // end method resetCnts.

    // ----

 // ####################################################################
 // ===========------- 	GRAMMAR-SPECIFIC METHODS ------=================
 // ####################################################################
 // ----
 // The following methods can be changed OR deleted as needed.  These
 // are grammar-specific methods that will be called by the
 // "loadTerminalData" and "checkTerminal" methods, and perhaps by others
 // as well.  If new methods are required for a specific type of grammar
 // or terminalData Object, then they should be ADDED HERE.
 //
 // Methods in this section only need to be here if they are referenced
 // somewhere else IN THIS FILE; otherwise, they can be removed.
 // ----

    /* ==========================================================
     *
     * method getValidAttributes():
     *
     * -------- */

    public Vector getValidAttributes(Vector constraints, String head,
	Vector subtypes) {

	Vector [] terms = (Vector[])(terminalData);
	if (constraints == null) {
	// no constraints; all attributes are valid.
	    return terms[1];
	} // end if.

	// Get all subtypes of the types in constraints, so that we
	// can do a full check.
	Vector fullConstraints =
	    Subtype.getAllSubtypesIn(constraints, subtypes);

	Vector options = new Vector();
	Iterator itr = terms[1].iterator();
	while (itr.hasNext()) {
	    String att = (String)itr.next();
	    if (fullConstraints.contains(getTypeOfTerminal(att, head))) {
	    // then this a valid choice.
		options.add(att);
	    } // end if.	    
	} // end while.

	return options;

    } // end method getValidAttributes.

    // ----

    /* ==========================================================
     *
     * method loadFuncs():
     *
     * -------- */

    public void loadFuncs() {

	this.funcMap.put(new Integer(this.funcMap.size() +1),
		"position");
	funcCnt++;
	this.funcMap.put(new Integer(this.funcMap.size() +1),
		"count");
	funcCnt++;
	this.funcMap.put(new Integer(this.funcMap.size() +1),
		"last");
	funcCnt++;
	this.funcMap.put(new Integer(this.funcMap.size() +1),
		"id");
	funcCnt++;
	this.funcMap.put(new Integer(this.funcMap.size() +1),
		"local-name");
	funcCnt++;
	this.funcMap.put(new Integer(this.funcMap.size() +1),
		"namespace-uri");
	funcCnt++;
	this.funcMap.put(new Integer(this.funcMap.size() +1),
		"name");
	funcCnt++;
	this.funcMap.put(new Integer(this.funcMap.size() +1),
		"concat");
	funcCnt++;
	this.funcMap.put(new Integer(this.funcMap.size() +1),
		"starts-with");
	funcCnt++;
	this.funcMap.put(new Integer(this.funcMap.size() +1),
		"contains");
	funcCnt++;
	this.funcMap.put(new Integer(this.funcMap.size() +1),
		"substring-before");
	funcCnt++;
	this.funcMap.put(new Integer(this.funcMap.size() +1),
		"substring-after");
	funcCnt++;
	this.funcMap.put(new Integer(this.funcMap.size() +1),
		"substring");
	funcCnt++;

	return;

    } // end method loadFuncs.

    // ----

    /* ==========================================================
     *
     * method loadSpaces():
     *
     * -------- */

    public void loadSpaces() {

	// Load a list of tokens before which we never
	// want a space.
	noSpaceBefore = new Vector();

	// Load a list of tokens after which we never want
	// a space.
	noSpaceAfter = new Vector();
	noSpaceAfter.add("\"$\"");
	noSpaceAfter.add("\"/\"");
	noSpaceAfter.add("\"//\"");

    } // end method loadSpaces.

    // ----

    public boolean isPunctuation(String str) {

	if (str.equals("\")\"") ||
	    str.equals("\"(\"") ||
	    str.equals("\"]\"") ||
	    str.equals("\"[\"") ||
	    str.equals("\"}\"") ||
	    str.equals("\"{\"") ||
	    str.equals("\"@\"") ||
	    str.equals("\":\"") ||
	    str.equals("\"::\"") ||
//	    str.equals("\"*\"") ||
//	    str.equals("\"+\"") ||
//	    str.equals("\"-\"") ||
	    str.equals("\".\"") ||
	    str.equals("\"..\"") ||
	    str.equals("\",\"") ||
	    str.equals("\"</\"") ||
	    str.equals("\"|\"") ||
	    str.equals("\" \"")) {
	    return true;
	} // end if.

	return false;

    } // end method isPunctuation.

    // ----

    public String getNewVarName() {

	String val = "$var" + varCnt;
	varCnt++;
	return val;

    } // end method getNewVarName.

    // ----

    public String getGenericId() {

	String val = "id" + genIdCnt;
	genIdCnt++;
	return val;

    } // end method getNewVarName.

    // ----

    public String getNewEltName() {

    // Create a new element name.

	String val = "newElt" + eltCnt;
	eltCnt++;
	return val;

    } // end method getNewEltName.

    // ----

    public String getNewAttrName() {

    // Create a new attribute name.

	String val = "newAttr" + attrCnt;
	attrCnt++;
	return val;

    } // end method getNewAttrName.

    // ----

    public String getNewFuncName() {

    // Create a new function name.  Add it's name to a map
    // so we can use it later.

	String val = "defFunc" + funcCnt;
	this.funcMap.put(new Integer(this.funcMap.size() +1), val);
	funcCnt++;
	return val;

    } // end method getNewFuncName.

    // ----

    public String getExistingFuncName() {

	if (this.funcMap.isEmpty()) {
	    String val = "undefFunc" + funcCnt;
	    funcCnt++;
	    return val;
	} // end if.
	else {
	    int funcKey = random.nextInt( this.funcMap.size()) + 1;
	    String val = (String) this.funcMap.get(new Integer(funcKey));
	    return val;
	} // end else.

    } // end method getExistingFuncName.

    // ----

    public String getCDATA() {

	// Determine a length.
	int length = random.nextInt(MAX_CDATA_LEN - 1) + 1;

	int tmp;
	String result = "";
	for (int i = 0; i < length; i++) {
	    tmp = random.nextInt(93) + 33;
	    while ((tmp == 34) || (tmp == 39)) {
	    // don't generate single or double quotes; they can mess
	    // things up.
		tmp = random.nextInt(93) + 33;
	    } // end if.
	    result += (char)(tmp);
	} // end for.

	return result;	

    } // end method getCDATA.

} // end class Generator.
