import java.io.*;
import java.lang.Math;
import java.util.*;

// -----------------============== NOTE =============--------------------
// To make changes to "terminal" data (for grammars other than XQuery),
// go to the "Generator.java" file--that is the ONLY file that needs to
// be modified for those purposes; everything else is designed to run
// independently of the grammar being used.
// -----------------=================================--------------------

// **********************************************************************
// GrammarNode class (base class for all other kinds of nodes).
// **********************************************************************

public class GrammarNode {

    // ********--========= Class variables ============--*********

    protected static int generationSeed;   // Seed for Generator object.
    protected static Generator generator;  // Used to generate specific id's,
					   //  variables, function names, etc.
					   //  when needed.

    RuleListNode myRuleList;		// Parsed grammar document.
    public static Vector rules;		// A vector to store the rules.
    public static int ruleCount;	// Number of rules (i.e. heads)
					//  in the grammar.

    public static Vector gramTypeDecls;	// Vector of type specs for the
					//  grammar.

    public static Vector gramOpTables;	// Vector of operation type tables
					//  for the grammar.

    public static Vector gramSubTypes;	// Vector of type equivalencies.

    protected static Map configMap;	// For holding configuration parameters.
    protected static int complexity;	// How "complex" the result should be.
    protected static int currDepth;	// # Of rules we've already evaluated.
    protected static int maxDepth;	// Maximum # of rules to evaluate before
					//  ending the generation.

    protected static String prevNode;	// Previous terminal string generated
					//  (for spacing purposes).

    // Possible components of a rule.
    protected static final int DEF_LIST = 0;
    protected static final int CHARSET = 1;
    protected static final int STRING_LIT = 2;
    protected static final int ID = 3;

    // Insertion options for vectors.
    protected static final int FRONT = 0;
    protected static final int REAR = 1;

    // Other constants.
    protected static final int MAX_COMPLEXITY = 500;
    protected static final int RECUR_FACTOR = 30;
    protected static final int DEFAULT_MAX_DEPTH = 20;
    protected static final int DEFAULT_SEED = 0;

    protected static final double DEFAULT_RECUR_PROB = 0.50;
    protected static final int DEFAULT_SYNTAX_WEIGHT = 1;
    protected static final int DEFAULT_TERM_SYNTAX = -1;

    protected static final int START_FRESH = -1;
    protected static final int NON_NESTED_RECUR = 0;
    protected static final char NO_RECURRENCE = '#';

    // Kinds of type declarations.
    protected static final boolean DYNAMIC = true;
    protected static final boolean NOT_DYNAMIC = false;

    // Kinds of type constraints.
    protected static final int REGULAR = 0;
    protected static final int CONDITIONAL = 1;

    // *******--========= End Class variables ============--*******

    /* ==========================================================
     *
     * Constructors.
     *
     * -------- */

    public GrammarNode() {}	// Default needed for subclasses.

    public GrammarNode(RuleListNode rulesList) {

	generator = null;
	myRuleList = rulesList;
	rules = new Vector();
	gramTypeDecls = new Vector();
	gramOpTables = new Vector();
	gramSubTypes = new Vector();
	configMap = new HashMap();
	currDepth = 0;
	complexity = 1;
	prevNode = null;

    } // end constructor.

    // ----

    /* ==========================================================
     *
     * method unparse():
     *
     * Build and process this GrammarNode structure based on inputs
     * and gathered stats, and then echo the structure to file.
     *
     * -------- */

    public void unparse(PrintWriter p, int indent) {

        myRuleList.unparse(p, 0);

    } // end method unparse.

    // ----

    /* ==========================================================
     *
     * method loadConfigFile():
     *
     * Reads the configuration file and builds a map to store
     * all of the necessary recurrence probabilities and syntax
     * weights.  Also loads the maxDepth and Seed values from
     * the config file.
     *
     * -------- */

    public void loadConfigFile(FileReader configFile,
		String complex) {

	System.out.println("Loading config file...");

	// Set the complexity.
	try {
	    setComplexity(Integer.valueOf(complex).intValue());
	} catch (NumberFormatException ex) {
	    System.err.println("\n-=- ERROR: Invalid complexity.\n" +
		"\nUsage: java gramParse <grammar_file> " +
		"<result_file> <complexity> [-o(verwrite)]\n" +
		"--> 'complexity' should be an integer between 1 and 500.\n");
	    System.exit(-1);
	} // end catch.

	// Temp vars used for processing.
	String delims = "\t ";
	StringTokenizer tokenizer;
	String name;
	double value;

	int lineCount = 0;
	String config = readLine(configFile, '\n');
	while (!config.equals("EOF")) {
	    lineCount++;
	    tokenizer = new StringTokenizer(config, delims);
	    if (tokenizer.hasMoreTokens()) {
		name = tokenizer.nextToken();
		if (name.equals("//")) {
		// then ignore this line.
		    config = readLine(configFile, '\n');
		    continue;
		} // end if.
		else if ((name.equals("SEED")) && tokenizer.hasMoreTokens()) {
		// Load the seed.
		    generationSeed = (Integer.valueOf
				(tokenizer.nextToken()).intValue());
		} // end else if.
		else {
		    if (tokenizer.hasMoreTokens()) {
			value = Double.valueOf(tokenizer.nextToken()).doubleValue();
			// Bump up recurrence values according to complexity.
			if (name.indexOf("RECUR") != -1) {
			    if (name.indexOf("MAX") == -1) {
			    // probability of recurrence.
				value += 1.0*complexity/MAX_COMPLEXITY;
			    } // end if.
			    else {
			    // max # of recurrences.
				value += complexity/RECUR_FACTOR;
			    } // end else.
			} // end if.
			if (configMap.containsKey(name)) {
			// duplicate.
			    System.out.println("-=- ERROR: Duplicate definition"
				+ " for head \"" + name + "\".\n");
			} // end if.
			else {
			    configMap.put(name, new Double(value));
			} // end inner else.
		    } // end inner if.
		    else {
		    // Invalid configuration line.
			System.err.println("\n-=- ERROR: Incomplete configuration" 
				+ "parameter (line " + lineCount + ").\n");
		    } // end inner else.
		} // end else.
	    } // end outer if.
	    config = readLine(configFile, '\n');
	} // end while.

	try { configFile.close(); }
	catch (Exception e) { e.printStackTrace(); }

	// Now that we have loaded the configuration parameters,
	// compute the syntax probability sums needed for random
	// generation.
	myRuleList.sumUpSyntaxes();

	// Finally, store the parameter that tells us what our maximum depth is.
//	maxDepth = (int)((Double)configMap.get("MAX_DEPTH")).doubleValue();
//	maxDepth *= Math.pow(complexity,2);	// depth = depth * complexity^2
	maxDepth = 10*complexity;

	System.out.println("--> Config File Loaded. <--\n");
	return;

    } // end method loadConfigFile.

    // ----

    /* ==========================================================
     *
     * method createConfigFile():
     *
     * Creates a configuration file for the grammar in question,
     * plugging default values into all the correct places and
     * leaving it up to the user to modify the configuration file
     * as needed for specific kinds of language generations.
     *
     * -------- */

    public void createConfigFile(PrintWriter p) {

	p.println("" +
	"\n// This file was automatically created by the RAGX generator.\n//\n" +
	"// For a given string \"ID_RECUR_X_Y,\" X is the syntax number within\n" +
	"// the rule that has ID as its head, and Y is the id of the occurence\n" +
	"// indicator within that syntax.\n" +
	"// Thus:\n" +
	"//              FAKE_SYNTAX_1\n" +
	"//              FAKE_RECUR_1_1\n" +
	"//              FAKE_RECUR_1_2\n" +
	"//              FAKE_SYNTAX_2\n" +
	"//              FAKE_RECUR_2_2\n" +
	"//\n" +
	"// describes a rule \"Fake\" that has 2 different syntaxes; of these\n" +
	"// syntaxes, the first has 2 occurence indicators, while the second\n" +
	"// only has one.  NOTE:  Occurrence indicators are numbered in the\n" +
	"// ORDER OF EVALUATION; thus, RECUR_1_1 refers to the indicator that\n" +
	"// will be evaluated first, RECUR_1_2 to the one that will be\n" +
	"// evaluated second, etc.\n" +
	"//\n" +
	"// For rules with more than one syntax choice, the final parameter is\n"+
	"// a 'termination syntax' parameter that ends with the word 'ENDING'\n" +
	"// (for exampe, FAKE_ENDING).  This parameter is for those cases where\n" +
	"// query generation results in Stack Overflow; it is used to specify\n" +
	"// which syntax of a given rule should be chosen when the generator\n" +
	"// has reached its maximum depth (a limit that is also set in this\n" + 
	"// file, presumably so that it is reached before stack overflow).\n" +
	"//\n" +
	"// The termination syntax should, in theory, be whatever choice leads\n" +
	"// to the nearest terminal, though this is in NO WAY checked by the\n" +
	"// generator (it is up to the person who modifies this file to make\n" +
	"// the appropriate termination syntax choices).  The default termination\n" +
	"// choice is '-1', which is a value that is ignored by the generator.\n" +
	"// For rules where the choice of syntax doesn't matter, the termination\n" +
	"// syntax should be set to '0'.\n");

	p.println("\nMAX_DEPTH\t" + DEFAULT_MAX_DEPTH);
	p.println("SEED\t\t" + DEFAULT_SEED + "\n\n");
        myRuleList.createConfigFile(p);

    } // end method createConfigFile.

    // ----

    /* ==========================================================
     *
     * method loadTypesFile():
     *
     * Reads the file containing type specifications for the
     * grammar, and stores those types so they can be used during
     * generation.
     *
     * -------- */

    public void loadTypesFile(FileReader typesFile) {

	System.out.println("Loading Types File...");

	// Temp vars used for processing.
	String delims = "\t ,:";
	StringTokenizer tokenizer;
	String head = null;

	Subtype tmpSubtype = null;
	TypeDecl tmpTDec = null;
	String typeDecString = readLine(typesFile, '@');
	while (!typeDecString.equals("EOF")) {
	    typeDecString = typeDecString.trim();
	    tokenizer = new StringTokenizer(typeDecString, delims);
	    if (tokenizer.hasMoreTokens()) {
		head = tokenizer.nextToken();
		head = head.trim();
		if (head.equals("SUBTYPES_OF")) {
		// then we have to load type equivalencies.
		    tmpSubtype = new Subtype(typeDecString);
		    if (tmpSubtype.getBase().equals("INVALID")) {
		    // error somewhere; abort.
			System.out.println("\n");
			System.exit(-1);
		    } // end if.
		    gramSubTypes.add(tmpSubtype);
		} // end if.
		else if (!head.equals("//")) {
		// create a typeDecl for the received input,
		// then store it in our running list.
		    tmpTDec = new TypeDecl(typeDecString);
		    if ((tmpTDec.getHead()).equals("INVALID")) {
			System.out.println("\n"); 
			System.exit(-1);
		    } // end if.
		    gramTypeDecls.add(tmpTDec);
		} // end else.
	    } // end if.
	    // Move to next entry.
	    typeDecString = readLine(typesFile, '@');
	} // end while.

	System.out.println("--> Types File Loaded. <--\n");
//	System.out.println(gramTypeDecls);	// Debugging
//	System.out.println(gramSubTypes);	// Debugging

	try { typesFile.close(); }
	catch (Exception e) { e.printStackTrace(); }

	return;

    } // end method loadTypesFile.

    // ----

    /* ==========================================================
     *
     * method loadOpTables():
     *
     * Reads the file containing operation tables for the grammar,
     * and stores those tables so they can be used during generation.
     *
     * -------- */

    public void loadOpTables(FileReader opsFile) {

	System.out.println("Loading Operation Tables...");

	// Temp vars used for processing.
	String delims = "\t ,:";
	StringTokenizer tokenizer;
	String head = null;

	OpTable tmpOpTbl = null;
	String opString = readLine(opsFile, '.');
	while (!opString.equals("EOF")) {

	    opString = opString.trim();
	    tokenizer = new StringTokenizer(opString, delims);
	    if (tokenizer.hasMoreTokens()) {
		head = tokenizer.nextToken();
		head = head.trim();
		if (head.equals("//")) {
		// then ignore this line.
		    opString = readLine(opsFile, '\n');
		} // end inner if.
		else {
		// create an opTable for the received input,
		// then store it in our running list.
		    tmpOpTbl = new OpTable(opString, gramTypeDecls, REGULAR);
		    if ((tmpOpTbl.getHead()).equals("INVALID")) {
			System.out.println("\n"); 
			System.exit(-1);
		    } // end if.
		    gramOpTables.add(tmpOpTbl);
		    opString = readLine(opsFile, '.');
		} // end else.
	    } // end if.
	    else {
	    // blank line; ignore it.
		opString = readLine(opsFile, '.');
	    } // end else.

	} // end while.

	System.out.println("--> Operation Tables Loaded. <--\n");
//	System.out.println(gramOpTables);

	try { opsFile.close(); }
	catch (Exception e) { e.printStackTrace(); }

	return;

    } // end method loadOpTables.


    // ----
 
    /* ==========================================================
     *
     * method getOpTable():
     *
     * Takes a rule head and a vector of operation tables, searches
     * that vector for a table with that head, and returns the table.
     * Returns 'null' if no table exists for the received rule head.
     *
     * -------- */
 
    public static OpTable getOpTable(String head, Vector gramOpTables) {
 
	OpTable tmpOpTbl = null;
	Iterator itr = gramOpTables.iterator();
	while (itr.hasNext()) {
	tmpOpTbl = (OpTable)itr.next();
	    if ((tmpOpTbl.getHead()).equals(head)) {
	    // we found it; return.
		return tmpOpTbl;
	    } // end if.
	} // end while.

	// If we get here, we didn't find it.
	return null;

    } // end method getOpTable.
 
    // ----

    /* ==========================================================
     *
     * method readLine():
     *
     * Reads a line from the received FileReader and returns
     * it.  The end of a 'line' is determined by the received
     * character.
     *
     * -------- */

    public String readLine(FileReader aFile, char lineEnd) {

	String str = "";
	try {
	    int tmpChar = aFile.read();
	    if (tmpChar == -1) {
		return "EOF";
	    } // end if.
	    while ((tmpChar != -1) && ((char)tmpChar != lineEnd)) {
		str += (char)tmpChar;
		tmpChar = aFile.read();
	    } // end while.
	} // end try.
	catch (IOException e) {
	    e.printStackTrace();
	} // end catch.

	return str;

    } // end method readLine.

    // ----

    /* ==========================================================
     *
     * method getDefinitionFor():
     *
     * Searches and returns the RuleNode that has a head matching
     * the received string; if no such node exists, null is returned.
     *
     * -------- */

    public RuleNode getDefinitionFor(String head) {

	Iterator itr = rules.iterator();
	RuleNode tmpNode = null;
	while (itr.hasNext()) {
	   tmpNode = (RuleNode)itr.next();
	   if ((tmpNode.head()).equals(head)) {
	   // we found it.
		return tmpNode;
	   } // end if.
	} // end while.

	// If we get here, we didn't find it.
	return null;

    } // end method getDefinitionFor.

    // ----

    /* ==========================================================
     *
     * method getTypeDecl():
     *
     * Searches and returns the Type declaration for the rule with
     * a head matching the received string; if no such type declaration
     * exists, null is returned.

     * -------- */

    public TypeDecl getTypeDecl(String head) {

	TypeDecl tmpTDec = null;
	Iterator itr = gramTypeDecls.iterator();
	while (itr.hasNext()) {
	    tmpTDec = (TypeDecl)itr.next();
	    if ((tmpTDec.getHead()).equals(head)) {
	    // we found it; return.
		return tmpTDec;
	    } // end if.
	} // end while.

	// If we get here, we didn't find it.
	return null;

    } // end method getTypeDecl.

    // ----

    /* ==========================================================
     *
     * method resetCount():
     *
     * Resets some stat variables for the beginning of another
     * rule generation.
     *
     * -------- */

    public void resetCount() {

	currDepth = 0;
	generator.resetCnts();
	prevNode = null;

    } // end method resetCount.

    // ----

    /* ==========================================================
     *
     * method setComplexity():
     *
     * Set the complexity of the generation to the received value,
     * making sure the resultant value is always valid.
     *
     * -------- */

    public void setComplexity(int value) {

	if (value < 1) {
	// min of '1'.
	    complexity = 1;
	} // end if.
	else if (value > MAX_COMPLEXITY) {
	// max.
	    complexity = MAX_COMPLEXITY;
	} // end else if.
	else {
	    complexity = value;
	} // end else.

	return;

    } // end method setComplexity.

    // ----

    /* ==========================================================
     *
     * method setTerminals():
     *
     * Creates a new Generator object using the received 'terminals'
     * object, then initiates that Generator object's Random variable
     * and echoes the seed to the screen.  Note that the Object received
     * might vary depending on the type of data referred to by the
     * terminals.  See the "Generator.java" and "TokenListNode.java"
     * files for further descriptions of how exactly the "terminals"
     * are used in the Generator and in query generation.
     *
     * -------- */

    public void setTerminals(Object obj) {

	generator = new Generator(obj);
	int seed = generator.setSeed(generationSeed);
	generator.setRandom();

	System.out.println("\n************************\n" +
			   "** Seed = " + seed +
			   " **\n************************\n");
	return;

    } // end method setTerminals.

} // end class GrammarNode.
