import java.util.*;

/* ***************************************************************
 *
 * ==============================================
 * Grammar
 * ==============================================
 *
 * OpTableDecl     ::= NonTerminal " :: " (RowSpec ";")* RowSpec "."
 * RowSpec         ::= PostEvalRow | Conditional
 * PostEvalRow     ::= TokenMatchList ResultSpec
 * TokenMatchList  ::= (TokSpec ",")*
 * TokSpec         ::= "TOKEN " TokenId " IS " (TypeName | "NONE")
 * ResultSpec      ::= ResultKeword ResultType
 * ResultKeyword   ::= "ELSE_RESULT_IS" | "THEN_RESULT_IS"
 * ResultType      ::= TypeName | ("TOKEN " TokenId)
 * Conditional     ::= "CONDITIONAL" WhenClause ConstraintClause
 * WhenClause      ::= "AFTER_TOKEN " TokenId "CHECK_THAT:"
 * ConstraintClause ::= (ConstraintDecl "#")* ConstraintDecl
 * ConstraintDecl  ::= TokenMatchList ConstraintSpec
 * ConstraintSpec  ::= ResultKeyword (Constraint | "NO_OP")
 * Constraint      ::= "FORCE_TOKEN" TokenId ForceClause SyntaxList?
 * ForceClause     ::= ("TO_TYPE" TypeList) | ("TO_TOKEN" TokenId)
 * SyntaxList      ::= "FOR_SYNTAX" SyntaxId (" " SyntaxId)*
 * 
 * ************************************************************** */

class OpTable {

    private String myHead;
    private int myType;
    private Vector typeSpecs;	// of "OpRow".

    private static final String REG_ROW_SEPARATOR = ";";
    private static final String COND_ROW_SEPARATOR = "#";
    private static final int REGULAR = 0;
    private static final int CONDITIONAL = 1;

    // ----

    /* ==========================================================
     *
     * Constructor:
     *
     * Given a string that constitutes an OpTable, pull off the
     * head, then load the body (i.e. the different RowSpecs).  The
     * received string should conform to the following syntax:
     *
     * OpTableDecl     ::= NonTerminal " :: " (RowSpec ";")* RowSpec "."
     *
     * -------- */

    public OpTable(String opDec, Vector gramTypeDecls, int tableType) {

	// Load the head (NonTerminal)
	int tmpPos = opDec.indexOf("::");
	if (tmpPos < 0) {
	    System.out.println("-=- ERROR: Invalid Operation Table" +
		" Spec (no head). -=-");
	    myHead = "INVALID";
	    typeSpecs = new Vector();
	    return;
	} // end if.

	// Else initialize values, then go.
	myHead = opDec.substring(0, tmpPos);
	myHead = myHead.trim();
	myType = tableType;
	typeSpecs = new Vector();

	opDec = opDec.substring(tmpPos + 2, opDec.length());
	if (opDec.length() == 0) {
	    System.out.println("-=- ERROR: Invalid Operation Table" +
		" Spec (no body) for '" + myHead + "' -=-");
	    myHead = "INVALID";
	    typeSpecs = new Vector();
	    return;
	} // end if.    

	// Now, load the operation type spec.
	if (myType == CONDITIONAL) {
	    doLoad(opDec, COND_ROW_SEPARATOR, gramTypeDecls);
	} // end if.
	else {
	    doLoad(opDec, REG_ROW_SEPARATOR, gramTypeDecls);
	} // end else.

	// By the time we get here, 'typeSpecs' has been loaded.
	// we're done.
	return;

    } // end Constructor.

    // ----

    /* ==========================================================
     *
     * doLoad():
     * 
     * Take the received string and parse into the various RowSpecs
     * making up the OpTable.  If we encounter any parse errors
     * while doing so, print an error message and return.  The
     * received 'opDec' string should conform to the following syntax,
     * with the 'NonTerminal ::' part removed (since it was already
     * processed).
     *
     * OpTableDecl     ::= NonTerminal " :: " (RowSpec ";")* RowSpec "."
     *
     * -------- */

    public void doLoad(String opDec, String rowSeparator,
	Vector gramTypeDecls) {

	String delims = "\t ,";
	StringTokenizer tokenizer;
	String tmpStr;
	boolean okay = false;
	int currPos = 0;
	int tmpPos = opDec.indexOf(rowSeparator, currPos);
	while (currPos >= 0) {
	// loop until entire table, is processed, RowSpec by RowSpec.

	    if (tmpPos < 0) {
	    // then this is the final entry (RowSpec) in the table.
		tmpStr = opDec.substring(currPos+1, opDec.length());
	    } // end if.
	    else {
		tmpStr = opDec.substring(currPos+1, tmpPos);
	    } // end else.
	    tmpStr = tmpStr.trim();
	    tokenizer = new StringTokenizer(tmpStr, delims);
	    if (!tokenizer.hasMoreTokens()) {
		System.out.println("-=- ERROR: Invalid Operation Table" +
		    " Spec (empty row). -=-");
		myHead = "INVALID";
		typeSpecs = new Vector();
		return;
	    } // end if.
	    // We've parsed out a single RowSpec in the table; load it now.
	    okay = loadRow(tmpStr, delims, gramTypeDecls);
	    if (!okay) {
	    // error somewhere; reset and get out.
		myHead = "INVALID";
		typeSpecs = new Vector();
		return;
	    } // end if.

	    // else, continue.
	    currPos = tmpPos;
	    tmpPos = opDec.indexOf(rowSeparator, currPos+1);

	} // end while.

    } // end method doLoad.

    // ----

    /* ==========================================================
     *
     * method loadRow():
     *
     * Takes a row of specs, received as a String, and loads it
     * into an OpRow data structure, which is a series of token/type
     * pairs followed by a final resultant type that will be used
     * if all of those token/type pairs are achieved (in other words,
     * an OpRow is used after ALL tokens in a token list node have
     * been generated (with types), and allows us to determine what
     * the final, overall type of that token list node should be;
     * that's why it's called a 'PostEvalRow'). If the row is that
     * of a conditional type constraint, then it is handled
     * appropriately.
     *
     * The received 'specStr' string should conform to the following
     * syntax:
     *
     * RowSpec         ::= PostEvalRow | Conditional
     * PostEvalRow     ::= TokenMatchList ResultSpec
     * TokenMatchList  ::= (TokSpec ",")*
     * 
     * The method returns 'true' if the row specification
     * was correctly parsed and loaded; 'false' otherwise.
     * 
     * -------- */

    private boolean loadRow(String specStr, String delims,
	Vector gramTypeDecls) {

	// First, see if we have a conditional type constraint.
	if (specStr.indexOf("CONDITIONAL") == 0) {
	// then we've got a conditional type constraint; deal
	// with it.
	    loadConditional(specStr, gramTypeDecls);
	    return true;
	} // end if.

	StringTokenizer tokenizer;
	int currPos = 0;
	int tmpPos = specStr.indexOf(",", currPos);

	// Place to hold the current type spec.
	OpRow rowSpec = new OpRow();
	String tmpStr;
	boolean okay;
	while (currPos >= 0) {
	// loop until entire TokenMatchList is processed, TokSpec by
	// TokSpec.

	    if (tmpPos < 0) {
	    // then this is the final type spec (i.e. ResultSpec).
		tmpStr = specStr.substring(currPos, specStr.length());
	    } // end if.
	    else {
		tmpStr = specStr.substring(currPos, tmpPos);
	    } // end else.
	    tmpStr = tmpStr.trim();
	    tokenizer = new StringTokenizer(tmpStr, delims);
	    if (!tokenizer.hasMoreTokens()) {
		System.out.println("-=- ERROR: Invalid Operation Table" +
		    " Spec (empty spec). -=-");
		return false;
	    } // end if.
	    // We've got a token/type (TokSpec) pair as a string; load it.
	    okay = getEntry(tokenizer, tmpStr, rowSpec);
	    if (!okay) {
	    // error somewhere; reset and get out.
		return false;
	    } // end if.

	    // else, continue.
	    currPos = tmpPos;
	    tmpPos = specStr.indexOf(",", currPos+1);

	} // end while.

	// If we get here, we're okay.  Add the row just read to
	// our running list.
	typeSpecs.add(rowSpec);

	// Done.
	return true;

    } // end method loadRow.

    // ----

    /* ==========================================================
     *
     * method getEntry():
     *
     * Parse and load a token/type pair, i.e. TokSpec.  The
     * received 'strTok' should conform to one of the following
     * syntaxs:
     *
     * TokSpec         ::= "TOKEN " TokenId " IS " (TypeName | "NONE")
     *
     * OR (if 'myType' is REGULAR):
     *
     * ResultSpec      ::= ResultKeyword ResultType
     * ResultKeyword   ::= "ELSE_RESULT_IS" | "THEN_RESULT_IS"
     *
     * OR (if 'myType' is CONDITIONAL):
     *
     * ConstraintSpec  ::= ResultKeyword (Constraint | "NO_OP")
     *
     * -------- */

    private boolean getEntry(StringTokenizer strTok, String rowAsStr,
	OpRow row) {

	String tmpStr = strTok.nextToken();
	tmpStr.trim();
	// Get TOKEN keyword.
	if (tmpStr.equals("TOKEN")) {
	    // --= Then is a TokSpec =--
	    if (strTok.countTokens() < 3) {
		System.out.println("-=- ERROR: Incomplete Spec" +
		   " entry: " + strTok + " -=-");
		return false;
	    } // end if.
	    // Get TokenId.
	    Integer tokId = null;
	    try {
		tokId = Integer.valueOf(strTok.nextToken());
	    } // end try.
	    catch (NumberFormatException e) {
		System.out.println("-=- ERROR: Invalid Spec Entry" +
		   " (not a number). -=-");
		return false;
	    } // end catch.
	    tmpStr = strTok.nextToken();
	    tmpStr = tmpStr.trim();
	    // Get IS keyword.
	    if (!tmpStr.equals("IS")) {
		System.out.println("-=- ERROR: Invalid entry " +
		    "keyword: '" + tmpStr + "' where 'IS' was " +
		    "expected. -=-");
		return false;
	    } // end if.
	    // Get TypeName.
	    tmpStr = strTok.nextToken();
	    // Now, store the token spec that we have.
	    row.specs.add(new TokenType(tokId, tmpStr));
	    return true;
	} // end if.
	else if (tmpStr.equals("THEN_RESULT_IS") ||
	   tmpStr.equals("ELSE_RESULT_IS")) {
	    // --= Then this is either a ResultSpec or ConstraintSpec =--
	    if (!strTok.hasMoreTokens()) {
		System.out.println("-=- ERROR: Incomplete Spec" +
		   " entry: " + strTok + "-=-");
		return false;
	    } // end if.
	    // Get the result type, and store it, using '-1' as an
	    // indicator that this is the RESULT type.  Note that
	    // the following call also sets the "resultToken" field of
	    // "row", if needed.
	    int pos = rowAsStr.indexOf("RESULT_IS");
	    String tmp = rowAsStr.substring(pos+10, rowAsStr.length());
	    row.specs.add(new TokenType(new Integer(-1),
		resultTypeOf(strTok, tmp, row)));
	    return true;
	} // end else.
	else {
	    System.out.println("-=- ERROR: Invalid entry keyword: '" +
		tmpStr + "' where 'TOKEN'/'RESULT_IS' was expected . -=-");
	    return false;
	} // end else.

    } // end method getEntry.

    // ----

    /* ==========================================================
     *
     * method resultTypeOf():
     *
     * Determines the result type of a ResultSpec or ConstraintSpec
     * declaration.  If the result type depends on a token (which can
     * only happen if this a ResultSpec), then that token is stored in
     * the received 'rowSpec' object.  If this a ConstraintSpec
     * meaning that myType is CONDITIONAL), then the result type is a
     * string to be used as a constraint type, and so is just
     * returned as is.
     *
     * The received resultType tokenizer should conform to one of the
     * followin syntaxes:
     *
     * If myType is REGULAR, then:
     * ResultSpec      ::= ResultKeword ResultType
     * ResultType      ::= TypeName | ("TOKEN " TokenId)
     *
     * otherwise (myType is CONDITIONAL):
     * ConstraintSpec  ::= ResultKeyword (Constraint | "NO_OP")
     *
     * -------- */

    public String resultTypeOf(StringTokenizer resultType,
		String resultAsStr, OpRow rowSpec) {

	if (myType == CONDITIONAL) {
	// then the "result" is actually a type constraint declaration;
	// we need to grab the entire string and return it (it will be
	// dealt with properly at the end of the "loadRow" method).
	    return resultAsStr;
	} // end if.

	// Else, load the result of the row.
	String tmpStr = resultType.nextToken();
	if (tmpStr.equals("TOKEN")) {
	// then result type depends on some other token; figure
	// out which one.
	    rowSpec.resultToken = null;
	    try {
		rowSpec.resultToken = Integer.valueOf(resultType.nextToken());
	    } // end try.
	    catch (NumberFormatException e) {
		System.out.println("-=- ERROR: Invalid Spec Entry" +
		   " (not a number). -=-");
		return "NONE";
	    } // end catch.
	    // we've stored the token on which the current row depends; set
	    // an indicator for later use (by 'checkSpec' method below) so
	    // that we know to use this token value.
	    return "AS_TOKEN";
	} // end if.
	else {
	// result type is given explicitly; just load it.
	    return tmpStr;	// tmpStr => TypeName
	} // end else.
	
    } // end method resultTypeOf.

    // ----

    /* ==========================================================
     *
     * method loadConditional():
     *
     * If we get here, we have a string representing a conditional
     * constraint; load it as necessary.  The received 'condDec'
     * string should conform to the following syntax:
     *
     * Conditional     ::= "CONDITIONAL" WhenClause ConstraintClause
     * WhenClause      ::= "AFTER_TOKEN " TokenId "CHECK_THAT:"
     * ConstraintClause ::= (ConstraintDecl "#")* ConstraintDecl
     *
     * -------- */

    private void loadConditional(String condDec, Vector gramTypeDecls) {

	// First, get the "base" token--i.e. the token that, when
	// generated, allows us to determine what the conditional
	// constraints are.
	int pos = condDec.indexOf(":");
	if (pos < 0) {
	    System.out.println("-=- ERROR: Invalid Conditional Spec" +
		" (no base token). -=-");
	    return;
	} // end if.
	StringTokenizer strTok = new StringTokenizer(condDec, " \t\n");

	// Get "AFTER_TOKEN" keyword (from WhenClause)
	strTok.nextToken();	// get past "CONDITIONAL" word.
	if (!strTok.nextToken().equals("AFTER_TOKEN")) {
	    System.out.println("-=- ERROR: Invalid conditional spec" +
		" (no base token). -=-");
	    return;
	} // end if.
	// Load base token.
	Integer baseTok = null;
	try {
	    baseTok = Integer.valueOf(strTok.nextToken());
	} // end try.
	catch (NumberFormatException e) {
	    System.out.println("-=- ERROR: Invalid Conditional" +
		" Spec (base token is not a number). -=-");
	    return;
	} // end catch.
	// Get "CHECK_THAT:" keyword.
	if (!strTok.nextToken().equals("CHECK_THAT:")) {
	    System.out.println("-=- ERROR: Invalid conditional spec" +
		" (missing 'CHECK_THAT:' keyword). -=-");
	    return;
	} // end if.
	// Now, load the conditional constraints (i.e. ConstraintClause).
	condDec = condDec.substring(pos+1, condDec.length());
	condDec = condDec.trim();
	OpTable condTable = new OpTable(myHead + " :: " + condDec,
		gramTypeDecls, CONDITIONAL);
	storeConditional(condTable, gramTypeDecls, baseTok);

	return;
	
    } // end method loadConditional.

    // ----

    /* ==========================================================
     *
     * method storeConditional():
     *
     * Takes a conditional constraint table (i.e. an OpTable with
     * myType set to CONDITIONAL) and stores it as a constraint in
     * the TypeDecl that corresponds to the rule for which the
     * current OpTable (this) was defined.  That conditional
     * constraint can then be used when it's time for generation.
     *
     * -------- */

    public void storeConditional(OpTable condTable,
	Vector gramTypeDecls, Integer baseTok) {

	TypeDecl tDec;

	// Find the correct TypeDecl.
	Iterator itr = gramTypeDecls.iterator();
	while (itr.hasNext()) {
	    tDec = (TypeDecl)itr.next();
	    if (tDec.getHead().equals(myHead)) {
	    // this is the type decl we need; load the receivef
	    // conditional constraint into it.
		tDec.loadCondConstr(condTable, baseTok);
		return;
	    } // end if.
	} // end while.

	// If we get here, we couldn't find a type decl for
	// the rule with 'myHead'; have to create a new one.
	tDec = new TypeDecl(myHead + " ==> NO_OP .");

	// Now, load the conditional constraint into it.
	tDec.loadCondConstr(condTable, baseTok);
	gramTypeDecls.add(tDec);

	// Done.
	return;

    } // end method storeConditional.

    // ----

    /* ==========================================================
     *
     * method determineType():
     *
     * If we get here, then the received 'types' parameter is full
     * of the token types for some generated rule.  Look at these
     * types, find the FIRST row of our OpTable that matches
     * them, and return the result type associated with that row.
     *
     * NOTE: If myType is REGULAR, the result of this call will be
     * the name of a type; if myType is CONDITIONAL, the result will
     * actually be a Constraint Declaration (in the form of a
     * string).
     *
     * -------- */

    public String determineType(String [] types) {

	OpRow rowSpec;
	String resType;

	// Iterate through table until we find a matching row.
	Iterator itr = typeSpecs.iterator();
	while (itr.hasNext()) {
	    rowSpec = (OpRow)itr.next();
	    resType = checkSpec(rowSpec, types);
	    if (resType != null) {
	    // then we found the type; return it.
		return resType;
	    } // end if.
	} // end while.

	// If we get here, then we went through the entire Op Table and
	// found no match (and no default).  Assume type is "NONE".
	return "NONE";

    } // end method determineType.

    // ----

    /* ==========================================================
     *
     * method checkSpec():
     *
     * Takes a received OpRow and checks each token/type pair
     * (i.e. TokenType) in that row to see if holds for the received
     * types.  A token/type pair "holds" if types[token] == type.
     * If the row holds, then the result type associated with
     * it is returned (as a string); otherwise, null is returned.
     *
     * NOTE: If myType is REGULAR, the result of this call will be
     * the name of a type; if myType is CONDITIONAL, the result will
     * actually be a Constraint Declaration (in the form of a
     * string).
     *
     * -------- */

    public static String checkSpec(OpRow row, String [] types) {

	TokenType tType;
	int tokId;
	Iterator itr = row.specs.iterator();
	while (itr.hasNext()) {
	    tType = (TokenType)itr.next();
	    if (tType.tokenId == -1) {
	    // then either we've reached the end of this row and everything
	    // matches, or we need the default ("else") type; either way,
	    // return the resultant type.
		if (tType.typeVal.equals("AS_TOKEN")) {
		// then result type depends on some token; use it.
		    if (row.resultToken.intValue() < types.length) {
			return types[row.resultToken.intValue()];
		    } // end if.
		    else {
		    // target token doesn't exist for current syntax;
		    // assume no match.
			return null;
		    } // end else.
		} // end if.
		// Else, just return the type that we have.
		return tType.typeVal;
	    } // end if.
	    // Else, see if row is still a possibility.
	    if (tType.tokenId >= types.length) {
	    // we're beyond the length of the spec; assume no match.
		return null;
	    } // end if.
	    // else, see if we have a match.
	    if (!types[tType.tokenId].equals(tType.typeVal)) {
	    // then this row of the table doesn't apply.
		return null;
	    } // end if.
	    // Else, we matched this one up, so go again...
	} // end while.

	// If we get here, we matched everything up, but didn't
	// find a result type.
	System.out.println("-=- ERROR: No result type indicated in" +
	    " Op Table row: '" + row + "'. -=-");
	return null;

    } // end method checkSpec.

    // ----

    /* ==========================================================
     *
     * method matchingRow():
     *
     * Looks through this OpTable to see if it can find a row that
     * matches the recevied constraints.  "Matching" means that either
     * the type for the row is in the list of received constraints,
     * or else it (the row) depends on some other token that might
     * potentially match the constraints.  Note that the FIRST
     * matching row is the one returned.
     *
     * -------- */

    public OpRow matchingRow(Vector constraints) {

	Iterator itr = typeSpecs.iterator();
	OpRow rowSpec;
	String resStr;
	while (itr.hasNext()) {
	    rowSpec = (OpRow)itr.next();
	    resStr = rowSpec.getResult();
	    if (resStr == null) {
	    // something went wrong; just give up.
		return null;
	    } // end if.
	    // Else, see if the result corresponds to one of the
	    // incoming types.
	    if (constraints.contains(resStr) ||
		resStr.equals("AS_TOKEN")) {
		// then we have a 'matching' row; return it.
		    return rowSpec;
	    } // end if.
	} // end outer while.

	// If we get here, none of the rows in the table match the
	// incoming constraints.
	return null;

    } // end method matchingRow.

    // ----

    /* ==========================================================
     *
     * method getHead():
     *
     * -------- */

    public String getHead() {

	return myHead;

    } // end method getHead.

    // ----

    /* ==========================================================
     *
     * method toString():
     *
     * -------- */

    public String toString() {

	return ("\n" + myHead + " ::\n" + typeSpecs);

    } // end method toString.

} // end class OpTable.
