import java.io.*;
import java.lang.Math;
import java.util.*;

// -----------------============== NOTE =============--------------------
// To make changes to "terminal" data (for grammars other than XQuery),
// go to the "Generator.java" file--that is the ONLY file that needs to
// be modified for those purposes; everything else is designed to run
// independently of the grammar being used.
// -----------------=================================--------------------

class RuleNode extends GrammarNode {

    // -------------------------------------------------------
    // 			-=-=-=- RuleNode -=-=-=-
    // -------------------------------------------------------
    // A rule in a given grammar consists of a head and a tail,
    // separated by a "::=" symbol.  We represent such a rule
    // with the RuleNode object, which has two data structures:
    // a HeadNode for the head, and a DefListNode for the tail.
    // -------------------------------------------------------

    // ************--==== Class Variables ====--************

    // Head and "Tail" nodes.
    private HeadNode myHead;
    private DefListNode myDefs;

    // Syntax sums generated from the config file for purposes
    // of selecting one of the available syntax choices for
    // generation.  For some syntax i, the value at partialSynSum[i]
    // is the sum of the weights of all syntaxes numbered 1 thru i,
    // including the weight for syntax i.
    private int [] partialSynSum;

    // **********--==== End Class Variables ====--**********

    // ----

    /* ==========================================================
     *
     * Constructor.
     *
     * -------- */

    public RuleNode(HeadNode head, DefListNode defs) {

	myHead = head;
	myDefs = defs;

    } // end constructor.

    // ----

    /* ==========================================================
     *
     * method unparse():
     *
     * Build and process this RuleNode structure based on inputs
     * and gathered stats, and then echo the structure to file.
     *
     * -------- */

    public void unparse(PrintWriter p, int indent) {

	myHead.unparse(p, indent);
	myDefs.unparse(p, indent);
	rules.add(this);

    } // end method unparse.

    // ----

    /* ==========================================================
     *
     * method generate():
     *
     * Randomly picks one of the available syntaxes in "tail"
     * (myDefs), and makes the call to generate it (taking care
     * to stop once the maxDepth is reached).  Only syntaxes
     * that can potentially satisfy the received type constraints
     * (forcedTypes) are considerd "available".
     *
     * -------- */

    public GResult generate(int currPos, Vector forcedTypes) {

	// Count each new rule generation as another depth (this
	// is used to keep the generator from causing stack over-
	// flow by recursing too deeply).
	currDepth++;

	// Var to hold result.
	GResult genResult = new GResult();

	// Stat vars used in generation.
	Vector currChoices = new Vector();	// 'Available' syntaxes.
	int [] currSynSums;			// Weighted sums for avail
						//  syntaxes.

	// Check to see if our syntax choice is restricted in any way.
	int restricted = checkRestrictions(forcedTypes, currChoices);
	if (restricted < 0) {
	// then we ran into problems--no semantically valid choices
	// exist.  Return special "fail" value.
	    genResult.strVal = "FAILED_GENERATION";
	    return genResult;
	} // end if.
	else if (restricted == 0) {
	// then we don't have any restrictions; ALL syntaxes are valid.
	    currChoices = myDefs.allChoices();
	    currSynSums = partialSynSum;
	} // end if.
	else {
	// we DO have restrictions, and the syntaxes that conform to them
	// are now in 'currChoices'.  Use that to find sums.
	    currSynSums = myDefs.sumUpSyntaxes(head(), currChoices);
	} // end else.

	// Load generation stats.
	int probTotal = currSynSums[myDefs.numChoices()-1];

	// Now, we have a set of (potentially) VALID syntax choices; pick
	// one of these choices at random, and generate it.
	genResult = doGeneration(forcedTypes, currPos, currChoices,
		probTotal, currSynSums);

	if (genResult != null) {
	// generation was successful; return result.
	    return genResult;
	} // end if.

	// Else, we failed to generate a type-valid body for this rule;
	// backup to the parent and try again (by returning "fail" value).
	genResult = new GResult();
	genResult.strVal = "FAILED_GENERATION";
	return genResult;

    } // end method generate.

    // ----

    /* ==========================================================
     *
     * method doGeneration():
     *
     * Given our constraints and the list of syntaxes which can potentially
     * satsify them, pick a syntax, generate it, and look at the resultant
     * type.  If the type does NOT abide by the constraints, then remove
     * that syntax from our list, randomly pick a different one, and try
     * again.  Repeat the process until we get a successful (i.e.
     * type-correct) result.  If we are unable to do so (meaning all
     * syntaxes are tried and none is type-correct), then we return
     * the null value, which is treated appropriately by the caller.
     *
     * Input parameters are as follows:
     * --------------------------------
     *
     * -> forcedTypes: list of 'target' types, meaning that the syntax
     *	we generate must have a result type that is included in this list.
     *  If the list is empty, then the result type can be anything.
     * -> currPos: the tokenId corresponding to the position of this
     * 	rule's head in the parent rule's token list.  For example,
     *	if this rule's head is "ruleA", and the parent rule's token
     *	list is:
     *			"half" "masted" ruleA ruleB
     *			 (0)     (1)     (2)    (3)
     *  then currPos is "2", because ruleA corresponds to token 2
     * 	in the list (token counting begins as '0').
     * -> currChoices: list of the potentially valid syntax choices
     *	that we have (from which we will randomly pick one).
     * -> probTotal: the total sum of all weights for syntaxes listed
     *	in 'currChoices'; this is used for selecting a syntax at random,
     *  based on the weights.
     * -> currSynSums: an array of weighted sums, where each element
     *	currSynSums[i] is the sum of all weights for syntaxes 1 thru (i+1),
     *	including the weight for syntax (i+1).
     *
     * -------- */

     public GResult doGeneration(Vector forcedTypes, int currPos,
		Vector currChoices, int probTotal, int [] currSynSums) {

	// Var to hold result.
	GResult genResult = new GResult();

	// Get total number of syntax choices we have for this rule.
	int totChoices = myDefs.numChoices();

	StatusCheck statCheck;
	while (true) {

	    statCheck = checkSpecialCases(forcedTypes, currPos, currChoices,
		probTotal);
	    if (statCheck.genRes != null) {
	    // we found a special case.
		if (statCheck.genRes.strVal.equals("FAILED_GENERATION")) {
		// special case failed; try again.
		    currChoices.remove(new Integer(statCheck.choice));
		    if (currChoices.size() == 0) {
		    // we've tried all potential syntaxes, and all failed.
			break;
		    } // end if.
		    // Else, try again...
		    currSynSums = myDefs.sumUpSyntaxes(head(), currChoices);
		    probTotal = currSynSums[totChoices-1];
		    continue;
		} // end if.
		// Else, special case succeeded; return the result.
		return statCheck.genRes;
	    } // end inner if.

	    int select = -1;	// which syntax are we going to generate.
	    int tmp = generator.getRandom().nextInt(probTotal);
	    for (int i = 0; i < totChoices; i++) {
		if (tmp < currSynSums[i]) {
		// then this is the syntax that we want.
		    select = i+1;
		    // See if incoming constraints have been met yet.
		    TypeDecl tDec = getTypeDecl(head());
		    if ((tDec != null) && tDec.hasDepFor(select)) {
		    // then our constraints have NOT been fulfilled yet
		    // (we either chose a dependent syntax, or a syntax
		    // that depends on an Op Table); keep them around.
			genResult = myDefs.generate(i, START_FRESH,
				head(), START_FRESH, START_FRESH,
				forcedTypes);
		    } // end if.
		    else {
		    // we've chosen a syntax with a fixed type (i.e. type
		    // was explicitly declared); so we do NOT need to pass
		    // the constraints any further.
			genResult = myDefs.generate(i, START_FRESH,
			   head(), START_FRESH, START_FRESH, null);
		    } // end else.
		    if (!(genResult.strVal.equals("FAILED_GENERATION"))) {
		    // then we're semantically correct; return.
			return genResult;
		    } // end inner if.
		    // else, the syntax that we chose was not type-correct;
		    // get out of the for loop, and try again...
		    break;
		} // end if.
	    } // end for.

	    // If we get here, our choice of syntax was not semantically
	    // correct; remove it from our list, and try again.
	    currChoices.remove(new Integer(select));
	    if (currChoices.size() == 0) break;
	    currSynSums = myDefs.sumUpSyntaxes(head(), currChoices);
	    probTotal = currSynSums[totChoices-1];

	} // end while.

	// If we get here, NONE of the syntaxes were type-correct; we need
	// to back-up to the parent, pick a different syntax (within
	// the _parent_ now), and try again. ('null' indicates this).
	return null;

    } // end method doGeneration.

    // ----

    /* ==========================================================
     *
     * method checkRestrictions():
     *
     * Look at the target types (of which we must match at least
     * one), and load into 'currChoices' all of the syntax choices
     * that can potentially give us one of those targets.  Then,
     * return an indicator saying whether or not the target types
     * actually limit our choices (i.e. wether or not we even need to
     * USE the currChoices vector).
     *
     * Input parameters are as follows:
     * --------------------------------
     *
     * -> forcedTypes: list of 'target' types, meaning that the syntax
     *	we generate must have a result type that is included in this list.
     *  If the list is empty, then the result type can be anything.
     * -> currChoices: list of the potentially valid syntax choices
     *	that we have (from which we will randomly pick one).
     *
     * -------- */

    public int checkRestrictions(Vector forcedTypes,
	Vector currChoices) {

	if (forcedTypes == null) {
	// then we have NO restrictions; all syntaxes/types are
	// possible.
	    return 0;
	} // end if.

	// Else, check contents of 'forcedType'.
	TypeDecl tDec = getTypeDecl(head());
	if (tDec == null) {
	// then no TypeDecl exists for this head (so even though we
	// we want to restrict it, we can't because we don't know how).
	// SO, just assume NO restrictions.
	    return 0;
	} // end if.
	else {
	    currChoices.addAll(tDec.getValidChoices(forcedTypes,
		gramSubTypes));
	    if (currChoices.size() == 0) {
	    // NO choices conform to restrictions; return indicator.
		return -1;
	    } // end if.
	    else if (currChoices.contains(new Integer(-1))) {
	    // then ALL choices are valid; no restrictions.
		return 0;
	    } // end else if.
	} // end outer else.

	// If we get here, then we DO actually have restrictions (and
	// the corresponding syntax choices are returned via the
	// 'currChoices' parameter).
	return 1;

    } // end method checkRestrictions.

    // ----

    /* ==========================================================
     *
     * method checkSpecialCases():
     *
     * Check to see if we're dealing with any special cases when we
     * generate this rule.  There are three 'special' cases that we
     * look at: 1) If there's only once choice, just generate it now
     * and return the result (don't need to bother with randomness);
     * 2) if, due the configuration file, all the valid syntax choices
     * sum up to zero, then we've been "told" not to generate anything;
     * call this a "failure" and return; 3) if we've reached our
     * maximum generation depth and need to terminate, then we have to
     * do things differently...
     *
     * Input parameters are as follows:
     * --------------------------------
     *
     * -> forcedTypes: list of 'target' types, meaning that the syntax
     *	we generate must have a result type that is included in this list.
     *  If the list is empty, then the result type can be anything.
     * -> currPos: the tokenId corresponding to the position of this
     * 	rule's head in the parent rule's token list.  For example,
     *	if this rule's head is "ruleA", and the parent rule's token
     *	list is:
     *			"half" "masted" ruleA ruleB
     *			 (0)     (1)     (2)    (3)
     *  then currPos is "2", because ruleA corresponds to token 2
     * 	in the list (token counting begins as '0').
     * -> currChoices: list of the potentially valid syntax choices
     *	that we have (from which we will randomly pick one).
     * -> probTotal: the total sum of all weights for syntaxes listed
     *	in 'currChoices'; this is used for selecting a syntax at random,
     *  based on the weights.
     *
     * -------- */

    public StatusCheck checkSpecialCases(Vector forcedTypes, 
	int currPos, Vector currChoices, int probTotal) {

	StatusCheck result = new StatusCheck();
	result.genRes = new GResult();

	// Check for 'special' case: only once choice.
	if (myDefs.numChoices() < 2) {
	// then there's only once choice; just generate it.
	    Integer tmpInt = (Integer)(currChoices.elementAt(0));
	    TypeDecl tDec = getTypeDecl(head());
	    if ((tDec != null) && tDec.hasDepFor(1)) {
	    // then our constraints have NOT been fulfilled yet
	    // (we chose a dependent syntax, or else one that uses
	    // an Op Table); keep them around.
		result.genRes = myDefs.generate(0, START_FRESH, head(),
		    START_FRESH, START_FRESH, forcedTypes);
		result.choice = tmpInt.intValue();
		return result;
	    } // end inner if.
	    else {
	    // the only available syntax has a fixed (explicit) type,
	    // so, we do NOT need to pass the target types any further.
		result.genRes = myDefs.generate(0, START_FRESH, head(),
		    START_FRESH, START_FRESH, null);
		result.choice = tmpInt.intValue();
		return result;
	    } // end inner else.
	} // end if.
	else if (probTotal == -1) {
	// more than one choice, but they all add up to zero.  In this
	// case, we set the 'result.choice' field to be the first choice
	// in our list of syntaxes.  We need to do this because, when
	// the "failure" is processed, the 'bad' syntax will be removed
	// from the list, until the list is empty (if we don't return
	// a 'bad' syntax, nothing will be removed from the list, so we
	// will get stuck in an infinite loop [the list will never become
	// empty]).
	    result.genRes.strVal = "FAILED_GENERATION";
	    result.choice = ((Integer)currChoices.elementAt(0)).intValue();
	    return result;
	} // end else if.
	else {
	// Check for 'special' case: need to terminate.
	    return checkTermination(currPos, currChoices, forcedTypes);
	} // end else.

    } // end method checkSpecialCases.

    // ----

    /* ==========================================================
     *
     * method checkTermination():
     *
     * Find the designated "termination syntax" for this rule,
     * and generate that definition.  If no termination syntax is
     * specified, then print a message and return nothing.
     *
     * Input parameters are as follows:
     * --------------------------------
     *
     * -> currPos: the tokenId corresponding to the position of this
     * 	rule's head in the parent rule's token list.  For example,
     *	if this rule's head is "ruleA", and the parent rule's token
     *	list is:
     *			"half" "masted" ruleA ruleB
     *			 (0)     (1)     (2)    (3)
     *  then currPos is "2", because ruleA corresponds to token 2
     * 	in the list (token counting begins as '0').
     * -> currChoices: list of the potentially valid syntax choices
     *	that we have (from which we will randomly pick one).
     * -> forcedTypes: list of 'target' types, meaning that the syntax
     *	we generate must have a result type that is included in this list.
     *  If the list is empty, then the result type can be anything.
     *
     * -------- */

    public StatusCheck checkTermination(int currPos,
		Vector currChoices, Vector forcedTypes) {

	StatusCheck result = new StatusCheck();
	result.genRes = new GResult();

	// See if we're trying to end the generation.
	if ((currDepth > maxDepth) &&
	    (currPos == 0))	/* Only terminate if we just started this
				 * tokList; if we're in the middle of the
				 * tokList, then we MUST respect type
				 * constraints, and so we can't just blindly
				 * use the termination choices. */ {

	    int choice = myDefs.terminate(head());
	    if (choice < 0) {
	    // none specified.
		System.out.println("\n-=- Generation terminated. -=-\n" +
			"-=- [ Max depth reached, but no termination syntax" +
			" specified. ] -=-\n-=- [ See your configuration" +
			" file for further details. ] -=-\n");
		// No point in going on; just quit.
		System.exit(1);
	    } // end inner if.
	    else if (choice > 0) {
	    // we have a termination syntax; generate it.
		TypeDecl tDec = getTypeDecl(head());
		if ((tDec != null) && tDec.hasDepFor(choice)) {
		// then our constraints have NOT been fulfilled yet
		// (we chose a dependent syntax, or one that uses an
		// Op Table); keep them around.
		    result.genRes = myDefs.generate(choice-1, START_FRESH,
			head(), START_FRESH, START_FRESH, forcedTypes);
		} // end inner if.
		else {
		// The termination syntax has an explicit type, so we do
		// NOT need to pass the target types any further.
		    result.genRes = myDefs.generate(choice-1, START_FRESH,
			head(), START_FRESH, START_FRESH, null);
		} // end inner else.
		// Note: if termination syntax wasn't semantically correct,
		// result.genRes.strVal will equal "FAILED_GENERATION", so
		// when we return 'result', the error will be found and we
		// will back out like normal.  However, to make sure that
		// we don't get stuck in an infinite try-retry loop, we
		// need to remove something from "currChoices" vector (but
		// this happens only in the event of a failure).  So, we
		// just pick the first element to remove.
		result.choice = ((Integer)currChoices.elementAt(0)).intValue();
		return result;
	    } // end else if.

	    // Else if choice == 0 we don't care which syntax we take
	    // (they all end at the same 'time'); return 'null' to
	    // let the caller know this.
	    result.genRes = null;
	    return result;

	} // end if.

	// If we get here, then we don't want to terminate; return
	// 'null' to let the caller know this.
	result.genRes = null;
	return result;

    } // end method checkTermination.

    // ----

    /* ==========================================================
     *
     * [ Other Helper Functions ]:
     *
     * The remaining functions simply allow access to, and
     * manipulation of, this RuleNode's data and statistics.
     *
     * -------- */

    public void sumUpSyntaxes() {

    // Load sums from this rule's syntax weights.

	partialSynSum = myDefs.sumUpSyntaxes(head(), myDefs.allChoices());

    } // end method sumUpSyntaxes.

    // ----

    public void createConfigFile(PrintWriter p) {

	myDefs.createConfigFile(p,head());

    } // end method unparse.

    // ----

    public String head() {

	return myHead.toString();

    } // end method head.

} // end class RuleNode
