import java.io.*;
import java.lang.Math;
import java.util.*;

// -----------------============== NOTE =============--------------------
// To make changes to "terminal" data (for grammars other than XQuery),
// go to the "Generator.java" file--that is the ONLY file that needs to
// be modified for those purposes; everything else is designed to run
// independently of the grammar being used.
// -----------------=================================--------------------

class TokenListNode extends GrammarNode {

    // -------------------------------------------------------
    // 		-=-=-=- TokenListNode -=-=-=-
    // -------------------------------------------------------
    // For a given grammar rule, the tail consists of one or
    // more possible syntax choices that together define the
    // head.  Each of these syntaxes constitutes a list (vector)
    // of token values that, when read from left to right, forms
    // a valid definition for the head--thus, we store each
    // syntax as a TokenListNode object.  For example, in the
    // following rule,
    //
    // fake ::= "def1" | "{" def2 "}" | "def3" def4+
    //
    // there are three different TokenListNodes:
    //
    //	1) "def1"	// 1 token (string lit)
    //	2) "{" def2 "}"	// 3 tokens (2 string lits, 1 id)
    //	3) "def3" def4+	// 2 tokens (1 string lit, 1 id).
    //
    // Note that each INDIVIDUAL choice is a TokenListNode; the
    // entire LIST of choices together (for a single head) is
    // is represented by the DefListNode structure (see
    // "DefListNode.java").
    //
    // -------------------------------------------------------

    // ************--==== Object Variables ====--************

    // Vector of TokenNodes (forming the list).
    private Vector myTokens;

    // 'List' of all occurence indicators that appear in this
    // particular TokenListNode.
    private String recurSymbols;

    // Array of 'types' for this token list (one for each token
    // in the token list).
    private String [] myTypes;

    // Local copy of "TypeDecl", for dealing with dynamic
    // constraints.
    private TypeDecl localTDec;

    // **********--==== End Object Variables ====--**********

    // ----

    /* ==========================================================
     *
     * Constructor.
     *
     * -------- */

    public TokenListNode(Vector V) {

        myTokens = V;
	recurSymbols = "";

    } // end constructor.

    // ----

    /* ==========================================================
     *
     * method unparse():
     *
     * Build and process this TokenListNode structure based on inputs
     * and gathered stats, and then echo the structure to file.
     *
     * -------- */
 
    public void unparse(PrintWriter p, int indent) {

	String str = new String();
	boolean first = true;
	TokenNode tmpNode;
	DefListNode tmpList;
	Iterator itr = myTokens.iterator();

	int cnt = 0;
	while (itr.hasNext()) {
	    if (!first) {
		p.print(" ");
	    } // end if.
	    tmpNode = (TokenNode)itr.next();
            tmpNode.unparse(p,0);
	    if (tmpNode.recurs()) {
		if (tmpNode.type() == DEF_LIST) {
		    tmpList = tmpNode.getList();
		    recurSymbols += tmpList.getRecurs();
		} // end if.
		recurSymbols += tmpNode.recurType();
	    } // end if.
	    first = false;
	    cnt++;
        } // end while.

	return;

    } // end unparse.

    // ----

    /* ==========================================================
     *
     * method unfold():
     *
     * "Unfolds" a TokenListNode by finding all nested DefListNodes
     * (i.e. list of tokens separated by vertical bars) and expanding
     * them out into multiple TokenListNodes so that each TokenListNode
     * has only ONE syntax choice contained within it.
     *
     * Ex.   The single syntax choice (TokenListNode)
     *          "def1" (def2 | def3) "def4"
     *       when put in UNFOLDED form, becomes TWO syntax choices:
     *          "def1" (def2) "def4",
     *          "def1" (def3) "def4".
     *
     *   Thus, the number of syntax choices goes from 1 to 2,
     *   and ALL of the resultant TokenListNodes contain only one
     *   possible syntax.  Technically, '(def2)' and '(def3)' are
     *   still DefListNodes, but each only contains ONE possible
     *   syntax choice (as evident by the lack of vertical bars),
     *   so no further unfolding is required.
     *
     * -------- */

    public Vector unfold() {

	// Intermediary variables.
	TokenNode tmpNode;
	Vector result = new Vector();		// TokenListNodes.
	Vector partial = new Vector();		// TokenListNodes.
	Vector tmpVec = new Vector();		// TokenListNodes.
	Vector thusFar = new Vector();		// TokenNodes.
	DefListNode nextList;

	Iterator itr = myTokens.iterator();
	while (itr.hasNext()) {
	// for every token in the TokenListNode:

	    tmpNode = (TokenNode)itr.next();
	    if (tmpNode.type() == DEF_LIST) {
	    // then we have a nested DefList...

		nextList = tmpNode.getList();
		if (nextList.numChoices() < 2) {
		// then this IS nested, but there's only
		// ONE syntax; treat it like 'normal'.
		    thusFar.add(tmpNode);
		    continue;
		} // end if.

		// Else, we have a nested DefList and it
		// DOES have multiple choices; unfold it.

		// Distribute occurence indicator throughout
		// all nested defLists so they are preserved
		// after the unfold.
		nextList.distribute(tmpNode.recurType());

		// Insert any NON-defList tokens into each
		// syntax of the DefListNode, so they are
		// preserved.
		nextList.fullInsert(thusFar);

		// Now, retrieve all TokenListNodes that
		// occur within the nested DefListNode.
		partial.addAll(nextList.tokLists());

		// Do a cross-product to determine ALL
		// possible syntax combinations (for cases
		// where a TokenListNode has multiple
		// DefListNodes, each with multiple choices).
		tmpVec.addAll(cross(result,partial));

		// Clean up.
		thusFar.clear();
		partial.clear();
		result.clear();

		// Store new result, then go again.
		result.addAll(tmpVec);
		tmpVec.clear();

	    } // end if.
	    else {
	    // this Token is NOT a DefListNode, so we just
	    // want to preserve it in the unfolding process.
		thusFar.add(tmpNode);
	    } // end else.

	} // end while.

	// Add remainder of the tokens (those appearing AFTER
	// the final nested defList) so they are preserved.

	TokenListNode tmpList;
	if (result.size() == 0) {
	// then no unfolding took place; just return the 'thusFar'
	// pieces, if there are any.
	    tmpList = new TokenListNode(thusFar);
	    tmpList.updateStats();
	    result.add(tmpList);
	    return result;
	} // end if.

	// Else, unfolding did take place; add the 'thusFar' pieces
	// to all unfolded elements.
	itr = result.iterator();
	while (itr.hasNext()) {
	    tmpList = (TokenListNode)itr.next();
	    tmpList.insert(thusFar,REAR);
	    tmpList.updateStats();
	} // end while.

	return result;

    } // end method unfold.

    // ----

    /* ==========================================================
     *
     * method cross():
     *
     * Given 2 vectors of TokenListNodes, this method finds the
     * cross product of those vectors (i.e. all possible syntax
     * combinations that can be formed from those vectors), and
     * returns the result.  Note that order of the parameters
     * DOES MATTER: in terms of syntax, elements from v1 will
     * always appear BEFORE (on the left of) elements from v2.
     *
     * Ex.  Given v1 = (def0 "def1"), (def0 "def2")
     * 		  v2 = (def3), (def4),
     *
     *	    then this method will return a vector containing
     *      FOUR TokenListNodes:
     *
     * 		result[0] = def0 "def1" def3
     * 		result[1] = def0 "def1" def4
     * 		result[2] = def0 "def2" def3
     * 		result[3] = def0 "def2" def4
     *
     * -------- */

    public Vector cross(Vector v1, Vector v2) {

	Vector result = new Vector();	// TokenListNodes
	Vector tmpVec = new Vector();	// TokenNodes.

	Iterator itr1 = v1.iterator();
	Iterator itr2 = v2.iterator();
	TokenListNode tmpList1;
	TokenListNode tmpList2;
	TokenListNode tmpList3;

	if (v1.size() == 0) {
	// then just deal with v2.
	    return v2;
	} // end if.

	// Else, do the cross.

	while (itr1.hasNext()) {
	    tmpList1 = (TokenListNode)itr1.next();
	    while (itr2.hasNext()) {

		tmpList2 = (TokenListNode)itr2.next();
		tmpVec.addAll(tmpList1.getList());
		tmpVec.addAll(tmpList2.getList());
		tmpList3 = new TokenListNode(tmpVec);
		result.add(tmpList3);
		tmpVec = new Vector();

	    } // end inner while.
	    itr2 = v2.iterator();
	} // end outer while.

	return result;

    } // end method cross.

    // ----

    /* ==========================================================
     *
     * method updateStats():
     *
     * This method clears out the TokenListNode's recurrence
     * symbols and stats, and then re-builds them by looking at
     * each individual Token.  This is required because, in the
     * "unfold" method, we alter pre-existing TokenListNodes as
     * well as create NEW ones, and so we need to account for
     * the new occurrence indicators in both cases.  For the same
     * reason, we also need to update the 'type' stats for this
     * TokenListNode.
     *
     * -------- */

    public void updateStats() {

	TokenNode tmpNode;
	DefListNode tmpList;
	Iterator itr = myTokens.iterator();

	recurSymbols = "";	// object var.

	while (itr.hasNext()) {
	    tmpNode = (TokenNode)itr.next();

	    // Update recurrence stats.
	    if (tmpNode.recurs()) {
		if (tmpNode.recurType() != NO_RECURRENCE) {
		    recurSymbols += tmpNode.recurType();
		} // end if.
		if (tmpNode.type() == DEF_LIST) {
		    tmpList = tmpNode.getList();
		    recurSymbols += tmpList.getRecurs();
		} // end if.
	    } // end if.

        } // end while.

	return;

    } // end method updateStats.

    // ----

    /* ==========================================================
     *
     * method insert():
     *
     * Takes a vector of TokenNodes and adds all elements in
     * that vector to this TokenListNode's list of Tokens
     * ("myTokens").  Whether the insertion occurs at the front
     * or the back of "myTokens" is determined by the 'loc'
     * parameter.
     *
     * -------- */

    public void insert(Vector nodes, int loc) {

	if (loc == FRONT) {
	// add at FRONT.
	    myTokens.addAll(0,nodes);
	} // end if.
	else {
	// add at REAR.
	    myTokens.addAll(nodes);
	} // end else.

	return;

    } // end method insert.

    // ----

    /* ==========================================================
     *
     * method generate():
     *
     * Generates a random but valid string of text corresponding
     * to the syntax defined by this TokenListNode.  This generation
     * is done on a Token-by-Token basis, where each Token is
     * generated according to the occurrence indicator associated
     * with it (if it has one).  In cases where the Token is an
     * id or a nested DefListNode, then the generation is done
     * recursively to find a valid syntax for that particular id
     * or DefListNode.
     *
     * The parameters received are as follows (most are needed for
     * the sake of finding recurrence probabilites).
     *
     *	"head" => 	The id of the rule for which we are generating
     *			a syntax.
     *  "syntax" =>	The number corresponding to the syntax choice
     *			that we are generating, w.r.t to the DefListNode
     *			that is the IMMEDIATE parent of this
     *			TokenListNode (could be a nested DefListNode).
     *  "origSyntax" =>	The number corresponding to the syntax choice
     *			that we are generating, w.r.t. the ORIGINAL
     *			DefListNode (the one at the very TOP of the
     *			recursion, representing the tail of some rule
     *			in the grammar file).
     *  "occurID" =>	The number corresponding to the occurrence
     *			indicator that we are processing (w.r.t
     *			the original TokenListNode at the top of the
     *			recursion).
     *	"nestedListID"  If we are generating a nested token list,
     *			then this is the number corresponding to the
     *			nested list we're processing, with respect to
     *			the head of the ORIGINAL (top-level) TokenList.
     *	"constraints"	A list of 'constraints' on this token list,
     *			meaning that the final type of the token list
     *			must be one of the types given in the "constraints"
     *			vector (or a subtype of one those types).
     *
     * Ex. Assume we are generating text from the SECOND choice
     * of the following rule:
     *
     *		fake ::= def0
     *		      | "def1"? (def2 def3*) ("def4")+
     *
     * In this case, there are technically THREE DefListNodes:
     *
     *		1) def0					// ORIGINAL
     *		   | "def1"? (def2 def3*) ("def4")+
     *		2) (def2 def3*)				// Nested.
     *		3) ("def4")				// Nested.
     *
     * Note that, as a result of "unfolding", all NESTED DefListNodes
     * will always have exactly ONE syntax choice by the time we get
     * to this "generate" method.
     *
     * In the given example, 'head' will be "fake", 'origSyntax'
     * will be "2" (because it's the 2nd choice of the ORIGINAL
     * DefListNode), and occurID will start out at "0" and run up
     * to "2" (since the ORIGINAL TokenListNode has a total of 3
     * occurence indicators).  'syntax' will initially start out
     * as "2", as well, since the immediate parent of the
     * TokenListNode is, in fact, the original DefListNode.  Finally,
     * 'nestedListID' will start at -1 (i.e. NOT nested), and the
     * contents of 'constraints' will depend on typing information
     * for the rule 'fake' (found in the "types" file).
     *
     * Once we start evaluating the TokenlistNode inside the first
     * nested DefListNode (def2 def3*), 'syntax' goes to "1"
     * because the immediate parent is now the NESTED DefListNode,
     * which only has one syntax choice; 'nestedListID' will go to
     * "1" because we are now processing the first nested list (with
     * respect to the original syntax choice); 'origSyntax' stays at
     * "2", though, and will for the entire generation of the rule.
     * When we hit the token [ def3* ], occurID will be "1" because
     * this is the second occurrence indicator in the ORIGINAL
     * TokenListNode ("def1"? (def2 def3*) ("def4")+), and counting
     * starts at '0'...
     *
     * -------- */

    public GResult generate(String head, int syntax, int origSyntax,
	int occurID, int nestedListID, Vector constraints) {

	// Initialize all of our data.

	// Store types and localTDec, as they are, so we don't clobber
	// them by recursion.
	String [] typesWhenEntered = saveTypes();
	TypeDecl localTDWhenEntered =
		saveLocalTD(head, syntax, nestedListID);

	// Now, initialize our types.
	initTypes();

	// Initialize status variables.
	occurID = initCounter(occurID, 1);

	// The following tells us which nested list we are currently
	// processing, taking into account any nested lists in _this_
	// token list object that we've _already_ processed.  This is
	// different from "nestedListID" because the latter ONLY counts
	// the nested lists evaluated BEFORE we started on this token
	// list (and it doesn't change through the evaluation of this
	// token list's tokens).
	int currNestID = initCounter(nestedListID, 0);

	int tokenCnt = 0;
	RecurStats recur;
	String spaces = "";
	String expandStr = null;
	StringBuffer str = new StringBuffer();
	GResult res = new GResult();

	// Load correct syntax.
	if (origSyntax != START_FRESH) {
	// in this case, we are dealing with a TokenListNode that
	// that is inside a nested DefListNode--but we need to use
	// syntax choice of the _original_ DefListNode for purposes
	// of looking up recurrence probabilities, etc.
	    syntax = origSyntax;
	} // end if.

	// Then, initialize our local TypeDecl (for dynamic constraints).
	initLocalTD(head, syntax, nestedListID, constraints);

	// Now generate.
	TokenNode tmpNode;
	Iterator itr = myTokens.iterator();
	while (itr.hasNext()) {
	// For each Token in this TokenList...

	    tmpNode = (TokenNode)itr.next();
	    // Get current spacing.
	    spaces = generator.getSpacingBetween(prevNode,
			tmpNode.idString());
	    // Expand this token out.
	    if (!tmpNode.recurs(NON_NESTED_RECUR)) {
	    // just expand it once and that's it.
		expandStr = expand(tmpNode, head, occurID,
			syntax, tokenCnt, currNestID, nestedListID,
			spaces, constraints);
		if (expandStr.equals("FAILED_GENERATION")) {
		// abort.
		    return abort(typesWhenEntered, localTDWhenEntered);
		} // end if.
		// else, continue.
		str.append(expandStr);
	    } // end if.
	    else {
	    // Handle recurrence as appropriate.
		recur = buildRecurStats(tmpNode.recurType(), occurID,
			head, syntax);
		if (currDepth > maxDepth) {
		// then we don't want any more recursion; only do it
		// if we must (i.e. if still under min).
		    while (recur.count < recur.min) {
			expandStr = expand(tmpNode, head, occurID,
			    syntax, tokenCnt, currNestID, nestedListID,
			    spaces, constraints);
			if (expandStr.equals("FAILED_GENERATION")) {
			// abort.
			    return abort(typesWhenEntered, localTDWhenEntered);
			} // end if.
			str.append(expandStr);
			// Update recurrence count.
			recur.count++;
		    } // end while.
		    // Now that we have the min generation, just go on.
		    // Check for conditional constraints based on this token.
		    checkCondConstr(head, syntax, nestedListID, tokenCnt);
		    // Increment counters.
		    occurID++;
		    tokenCnt++;
		    continue;
		} // end if.
		// Else, generate while recurrence stats allow it.
		while ((recur.count < recur.min) ||
		  ((generator.getRandom().nextDouble() < recur.prob) &&
		  (recur.count < recur.max))) {
		    expandStr = expand(tmpNode, head, occurID,
			    syntax, tokenCnt, currNestID, nestedListID,
			    spaces, constraints);
		    if (expandStr.equals("FAILED_GENERATION")) {
		    // abort.
			return abort(typesWhenEntered, localTDWhenEntered);
		    } // end if.
		    // else, continue.
		    str.append(expandStr);
		    // Update recurrence count.
		    recur.count++;
		} // end while.
		if (tmpNode.type() == DEF_LIST) {
		// need to account for all occurrence indicators used
		// in the nested list.
		    occurID += (tmpNode.getList()).recurs();
		} // end if.
		// Now, account for the occurrence indicator just used.
		occurID++;
	    } // end else.
	    // Check for conditional constraints based on this token.
	    checkCondConstr(head, syntax, nestedListID, tokenCnt);
	    // Update counters.
	    tokenCnt++;
	    if (tmpNode.type() == DEF_LIST) {
	    // need to account for all def lists used in the nested list,
	    // plus this one.
		currNestID += (tmpNode.getList()).numNestedLists();
		currNestID++;
	    } // end if.

	} // end outer while.

	// At this point, we know the types of all tokens in this
	// token list; use them to figure out what the type of the
	// token list as a whole is (i.e. the type of 'head' with
	// respect to this particular instance of it); return the
	// result.
	return determineType(head, str.toString(), nestedListID,
		syntax, typesWhenEntered, localTDWhenEntered);

    } // end method generate.

    // ----

    /* ==========================================================
     *
     * method saveTypes():
     *
     * Get a copy of the object variable 'myTypes' and return it;
     * this copy is used to restore the original when and if it
     * (the original) is clobbered by recursion.
     *
     * -------- */

    public String [] saveTypes() {

	if (myTypes == null) {
	// we haven't initialized them yet; do so now.
	    initTypes();
	} // end if.

	// Now, copy myTypes.
	String [] result = new String[myTypes.length];
	for (int i = 0; i < myTypes.length; i++) {
	    result[i] = myTypes[i];
	} // end for.

	return result;

    } // end method saveTypes.

    // ----

    /* ==========================================================
     *
     * method saveLocalTD():
     *
     * Get a copy of the object variable 'localTDec' and return it;
     * this copy is used to restore the original when and if it
     * (the original) is clobbered by recursion.
     *
     * -------- */

    public TypeDecl saveLocalTD(String head, int syntax, int nestID) {

	if (localTDec == null) {
	// assume it's not intialized yet.
	    TypeDecl td = findTypeDecl(head, syntax, nestID);
	    if (td != null) {
		return (new TypeDecl(td));
	    } // end if.
	    else {
		return null;
	    } // end else.
	} // end if.

	// Else, copy localTDec.
	return (new TypeDecl(localTDec));

    } // end method saveLocalTD.

    // ----

    /* ==========================================================
     *
     * method initTypes():
     *
     * Initialize all values in "myTypes" (i.e. the types of all
     * tokens in this token list object) to "NONE".  After generation,
     * the ONLY types that will remain as "NONE" are those associated
     * with a Token that a) has an occurrence indicator tied to it,
     * and b) because of that occurrence indicator, was never actually
     * generated.  For ALL other tokens, if a type can't be found, it
     * is set to equal the head of the rule containing the token (it
     * does NOT stay as "NONE").
     *
     * -------- */

    public void initTypes() {

	// Initial type vector.
	myTypes = new String[myTokens.size()];
	for (int j = 0; j < myTypes.length; j++) {
	    myTypes[j] = "NONE";
	} // end for.

	// Done.
	return;

    } // end method initTypes.

    // ----

    /* ==========================================================
     *
     * method initLocalTD():
     *
     * Creates a local copy of the global TypeDecl for the rule
     * corresponding to the received head.  Then, it looks at
     * the target types ('constraints') to which this token list
     * must adhere, and adds any DYNAMIC token constraints to
     * the local TypeDecl to ensure that at least one of the
     * target types can be generated.  Dynamic constraints are
     * based on the "types" and "op_tables" input files; if no
     * entries exist in these files for the received head, or
     * if no target types are actually received in this call, then
     * no dynamic constraints are generated.
     *
     * -------- */

    public void initLocalTD (String head, int syntax, 
		int nestID, Vector constraints) {

	// Get the global TypeDecl for 'head'.
	TypeDecl tDec = findTypeDecl(head, syntax, nestID);

	// Create a local copy.
	if (tDec == null) {
	    localTDec = null;
	    // We can't do anything with incoming types if there is
	    // no type decl for this rule; nothing else to do.
	    return;
	} // end if.
	else {
	    // Get local copy.
	    localTDec = new TypeDecl(tDec);
	    // See if type of this token list depends on a specific
	    // token, and if so, force THAT token to be a valid type
	    // (by adding a dynamic constraint to 'localTDec').
	    boolean done = forceToIncomingType(head, syntax,
		nestID, constraints);
	    if (!done) {
	    // type does NOT depend on a specific token; see if it 
	    // relies on an op table.
	        if ((tDec.typeForSyntax(syntax) != null)
	  	   && tDec.typeForSyntax(syntax).equals("OP_TABLE")) {
		// if so, look at the op table and try to find a row that
		// gives us the correct type.  We will reflect the chosen
		// row by adding dynamic constraints to "localTDec".
		    processOpTableTypes(head, syntax, nestID, constraints);
		} // end inner-most if.
		// NOTE: if typeForSyntax(syntax) returns something other
		// than 'OP_TABLE', then it must be an explicit type, and
		// thus cannot be forced to anything else; so, we don't
		// do anything.
	    } // end inner if.
	} // end else.

	// Done.
	return;

    } // end method initLocalTD.

    // ----

    /* ==========================================================
     *
     * method forceToIncomingType():
     *
     * Checks to see if the type of the current token list (syntax)
     * depends on a specific token in the list; if so, then it
     * generates a dynamic constraint forcing _that_ token to adhere
     * to the target types ('constraints'), and adds that dynamic
     * constraint to the local copy of the TypeDecl.  Returns 'true'
     * if a token dependency exists and we were able to add a dynamic
     * constraint; returns 'false' otherwise.
     *
     * -------- */

    public boolean forceToIncomingType(String head, int syntax, int nestID,
		Vector constraints) {

	// First, make sure we have a valid local copy of the TypeDecl
	// corresponding to 'head'.
	if (localTDec == null) {
	// no type decl exists for this nested token list; we have no
	// choice but to ignore the constraint (which means the generation
	// will probably fail and a different path will be chosen).
	    return false;
	} // end if.

	// Else, we have a relevant type decl; check to see if the type
	// depends on a specific token.
	if (localTDec.tokDependency(syntax) != null) {
	// then we know the token on which the result type depends; now,
	// we have to add a new (dynamic) constraint forcing that token to
	// one of the target (incoming) types.
	    int targetTok = (localTDec.tokDependency(syntax)).intValue();
	    return addDynamicConstraint(syntax, constraints, targetTok);
	} // end if.

	// If we get here, the type of this token list does not depend on
	// a specific token; if it is determined by an Op Table, then we
	// will handle it later (in the 'initLocalTD' method).  If it is NOT
	// determined by an op table, then we can't do anything more;
	// just ignore the constraint (which means the generation will
	// probably fail and a different path will be chosen).
	return false;

    } // end method forceToIncomingType.

    // ----

    /* ==========================================================
     *
     * method addDynamicConstraint():
     *
     * Add a dynamic token constraint to the local TypeDecl object.
     * Do so by generating the constraint as a string, then making
     * a call to the localTDec to add the constraint to it's list.
     * Returns 'true' if we successfully added the constraint; returns
     * false otherwise.
     *
     * -------- */

    public boolean addDynamicConstraint(int syntax,
		Vector constraints, int targetTok) {

	if ((constraints == null) || (constraints.size() == 0)) {
	// if there aren't any constraints, don't bother doing anything.
	    return false;
	} // end if.

	// Else, add target types; note that "gramTypeDecls" comes from the
	// GrammarNode class, which is the parent of this one.

	// First, create the constraint as a string (see Type file
	// grammar for specifics on what the string has to look like).
	String dynConstr = targetTok + " TO_TYPE";
	Iterator itr = constraints.iterator();
	while (itr.hasNext()) {
	    dynConstr = dynConstr + " " + (String)itr.next();
	} // end while.
	dynConstr = dynConstr + " FOR_SYNTAX " + syntax;

	// Make the string into a string tokenizer.
	StringTokenizer str = new StringTokenizer(dynConstr);

	// Add it to the TypeDecl.
	localTDec.loadConstraint(str, DYNAMIC);

	// Now we're done; the dynamic constraint will be enforced
	// automatically through calls to the "expand" method.
	return true;

    } // end method addDynamicConstraint.

    // ----

    /* ==========================================================
     *
     * method processOpTableTypes():
     *
     * Look at the op table to find out what the different possible
     * result types are for the given head; then add token
     * constraints such that the resultant type will be one of the
     * target (incoming) constraints.
     *
     * -------- */

    public void processOpTableTypes(String head, int syntax,
		int nestID, Vector constraints) {

	// If we don't have any constraints, there's nothing to do.
	if ((constraints == null) || (constraints.size() == 0)) {
	    return;
	} // end if.

	// Figure out what our head is.
	if (nestID != START_FRESH) {
	    head = head + "_NESTED_" + syntax + "_" + nestID;
	} // end if.
	// else, head stays as it is.

	// First, get the appropriate op table; note that "gramOpTables"
	// comes from the GrammarNode class, which is the parent of this
	// one (likewise for the method "getOpTable").
	OpTable opT = getOpTable(head, gramOpTables);
	if (opT == null) {
	// no op table exists for this rule; nothing we can do; ignore
	// the incoming constraints (which means the generation
        // will probably fail and a different path will be chosen).
	    return;
	} // end if.

	// Else, run through the op table and find the FIRST row that
	// matches any of the incoming constraints.
	OpRow chosenRow = opT.matchingRow(constraints);
	if (chosenRow != null) {
	// then we found a row that matches; go through and create all of
	// the necessary token constraints to force this row to be chosen.
	    TokenType tType;
	    Iterator itr = chosenRow.specs.iterator();
	    while (itr.hasNext()) {
		tType = (TokenType)itr.next();
		if (tType.tokenId == -1) {
		// then this a result type; if the result type depends
		// on a token, then we have to add a constraint to force
		// _that_ token to be the correct type.
		    if (tType.typeVal.equals("AS_TOKEN")) {
			addDynamicConstraint(syntax, constraints,
			    chosenRow.resultToken.intValue());
		    } // end inner if.
		    break;
		} // end if.
		// Else, add a dynamic constraint.
		if (tType.typeVal.equals("NONE")) {
		// don't force anything to NONE; ignore this constraint.
		    continue;
		} // end if.
		String dynConstr = tType.tokenId + " TO_TYPE " + tType.typeVal;
		dynConstr = dynConstr + " FOR_SYNTAX " + syntax;
		StringTokenizer str = new StringTokenizer(dynConstr);
		localTDec.loadConstraint(str, DYNAMIC);
	    } // end while.
	    return;
	} // end if.

	// Else, no matching rows; the only possible way to adhere
	// to the incoming constraints is to force the "ELSE" clause to
	// be chosen, which can be difficult. So instead, we just
	// give up and ignore the constraints (which means the
	// generation will probably fail and a different path will
	// be chosen).
	return;

    } // end method processOpTableTypes.

    // ----

    /* ==========================================================
     *
     * method initCounter():
     *
     * Takes a variable and a default value; if the variable is
     * not set (i.e. if it equals START_FRESH), then the default
     * value is returned.  Otherwise, the variable itself is
     * returned.
     *
     * -------- */

    public int initCounter(int var, int defaultVal) {

	if (var == START_FRESH) {
	// then this TokenListNode is the 'original' one at the top
        // of the recursion; initialize the counter to the default.
	    return defaultVal;
	} // end if.
	else {
	// start it at whatever value we received.
	    return var;
	} // end else.

    } // end method initCounter.

    // ----

    /* ==========================================================
     *
     * method expand():
     *
     * Takes a TokenNode and returns the String corresponding to
     * it.  If the TokenNode is a DefListNode, then a recursive
     * generate call is made on that DefListNode; if it is an Id,
     * then a rule with that Id is found and generated recursively;
     * if no such rule exists, then the "terminals" object of the
     * Generator is searched.  In the event that no terminal
     * exists, either, then an error message is printed and the
     * TokenNode's id is simply returned.
     *
     * The parameters received are as follows:
     *
     *	"tmpNode" =>	The Token we are currently expanding.
     *	"head" => 	The id of the rule for which we are generating
     *			a syntax.
     *  "occurID" =>	The number corresponding to the occurrence
     *			indicator that we are processing (w.r.t
     *			the original TokenListNode at the top of the
     *			recursion).
     *  "syntax" =>	The number corresponding to the syntax choice
     *			that we are generating, w.r.t. the ORIGINAL
     *			DefListNode (the one at the very TOP of the
     *			recursion, representing the tail of some rule
     *			in the grammar file).
     *	"tokenCnt" =>	The number corresponding to the 'tmpNode' token
     *			w.r.t. the current syntax.
     *	"nestCnt" =>	If we are generating a nested token list, this
     *			the number corresponding to the nested list we're
     *			processing, taking into account any nested lists
     *			that we've already processed in _this_ token list
     *			object.
     *	"nestID" =>	If we are generating a nested token list,
     *			then this is the number corresponding to the
     *			nested list we're processing, with respect to
     *			the head of the ORIGINAL (top-level) TokenList.
     *	"spacing" =>	Any spacing that needs to occur between the previous
     *			token's expanded string and this token's expanded
     *			string.
     *	"constraints"	A list of 'constraints' on this token list,
     *			meaning that the final type of the token list
     *			must be one of the types given in the "constraints"
     *			vector (or a subtype of one those types).
     *
     * -------- */

    public String expand(TokenNode tmpNode, String head,
		int occurID, int syntax, int tokenCnt,
		int nestCnt, int nestID, String spacing,
		Vector constraints) {

	if (tmpNode.type() == STRING_LIT) {
	// then we just keep the token as is, but remove quotes.
	    String str = tmpNode.idString();
	    prevNode = str;
	    myTypes[tokenCnt] = "TERMINAL";
	    return (spacing + str.substring(1,str.length()-1));
	} // end if.
	else if (tmpNode.type() == CHARSET) {
	// generate one of the characters in the set.
	    String selection = tmpNode.charSelect();
	    prevNode = selection;
	    myTypes[tokenCnt] = "TERMINAL";
	    return selection;
	} // end else if.
	else if (tmpNode.type() == DEF_LIST) {
	// Recurses.  By this time, all rules have been "unfolded", so
	// we know there's only one valid syntax for this particular
	// def list; it's id is '0'.  Generate it.
	    GResult res;
	    Vector forcedTypes = prepForRecursion(head,
			tokenCnt, syntax, nestID);
	    DefListNode nextList = tmpNode.getList();
	    if (tmpNode.recurs(NON_NESTED_RECUR)) {
		res = nextList.generate(0, syntax, head,
			occurID + 1, nestCnt + 1, forcedTypes);
	    } // end if.
	    else {
		res = nextList.generate(0, syntax, head,
			occurID, nestCnt + 1, forcedTypes);
	    } // end else.
	    if (myTypes[tokenCnt].equals("NONE")) {
		myTypes[tokenCnt] = res.typeVal;
	    } // end if.
	    return res.strVal;
	} // end else if.
	else {
	// We have an 'id', which means that we need to
	// generate the tail of some rule with a head
	// matching 'id'.
	    RuleNode currRule = getDefinitionFor(tmpNode.idString());
	    if (currRule != null) {
	    // then we found a matching rule; recurse.
		Vector forcedTypes = prepForRecursion(head,
			tokenCnt, syntax, nestID);
		GResult genResult = currRule.generate(tokenCnt, forcedTypes);
		if (myTypes[tokenCnt].equals("NONE")) {
		    if (genResult.typeVal != null) {
			myTypes[tokenCnt] = genResult.typeVal;
		    } // end if.
		} // end if.
		return genResult.strVal;
	    } // end if.
	    else {
		// Check to see if this head requires 'terminal'
		// expansion (for example, in the case of XQuery
		// generation, check to see if it should return some
		// element or attribute from the parsed DTD).
		String term = generator.checkTerminal(tmpNode.idString(),
				constraints, gramSubTypes);
		if (term.equals("NONE")) {
		// we couldn't find anything; just return the id.
		    System.out.println("\n-=- NOTE: Couldn't find rule " +
			"for \"" + tmpNode.idString() + "\". -=-");
		    return (spacing + tmpNode.idString());
		} // end if.
		else if (term.equals("INVALID")) {
		// terminal exists, but isn't of the correct type.
		    return "FAILED_GENERATION";
		} // end else if.
		else {
		// else we found one; store its type, then return it.
		   myTypes[tokenCnt] = generator.getTypeOfTerminal(term,
			tmpNode.idString());
		   prevNode = generator.getValueOfTerminal(term,
                        tmpNode.idString());
		   return (spacing + generator.getValueOfTerminal(term,
			tmpNode.idString()));
		} // end if.
	    } // end inner else.
	} // end else.

    } // end method expand.

    // ----

    /* ==========================================================
     *
     * method prepForRecursion():
     *
     * Looks at what the token constraints are for this token
     * list object, then figures out what types the current token
     * can be in order to satsify those constraints.  The list of
     * of valid ("target") types is returned as a vector.
     *
     * -------- */

    public Vector prepForRecursion(String head, int tokenCnt,
		int syntax, int nestID) {

	Vector forcedTypes;
	if (localTDec == null) {
	    forcedTypes = null;
	} // end if.
	else {
	    forcedTypes = localTDec.getConstraints(tokenCnt,
		myTypes, syntax);
	} // end else.

	// Done.

	return forcedTypes;

    } // end method prepForRecursion.

    // ----

    /* ==========================================================
     *
     * method buildRecurStats():
     *
     * Loads the recurrence statistics for the received occurence
     * indicator, according to the values read in from the
     * configuration file.  These stats are returned via the
     * "RecurStats" data structure.
     *
     * -------- */

    public RecurStats buildRecurStats(char indic, int occurCnt,
			String head, int syntax) {

	// Initialize recurrence stat variables.
	RecurStats result = new RecurStats();

	// Now build the statistics.
	String recurID;
	recurID = head.toUpperCase() + "_RECUR_" + 
		syntax + "_" + occurCnt;

	result.prob = lookup(recurID);
	result.indicator = indic;
	result.count = 0;

	switch(indic) {
	    case '*':
		result.max = (int)lookup(recurID + "_MAX");
		result.min = 0;
		break;
	    case '+':
		result.max = (int)lookup(recurID + "_MAX");
		result.min = 1;
		break;
	    case '?':
		result.max = 1;
		result.min = 0;
		break;
	    default:
		System.out.println("\n-=- ERROR: Unexpected occurence" +
			" indicator: " + indic + "\n");
	} // end switch.

	return result;

    } // end method buildRecurStats.

    // ----

    /* ==========================================================
     *
     * method abort():
     *
     * If we get here, we have a case of a failed generation (meaning
     * some typing constraints were not met). Restore local variables, 
     * then return a "failure" indication.
     *
     * -------- */

    public GResult abort(String [] entryTypes, TypeDecl entryLocalTD) {

	restoreTypes(entryTypes);
	restoreLocalTD(entryLocalTD);

	GResult res = new GResult();
	res.strVal = "FAILED_GENERATION";
	return res;

    } // end method abort.

    // ----

    /* ==========================================================
     *
     * method checkCondConstr():
     *
     * See if there are any conditional constraints for which the
     * base token matches the one received; if so, we need to add
     * the correct constraints (dynamically), as indicated by the
     * conditional description in the Op Table.
     *
     * -------- */

    private void checkCondConstr(String head, int syntax,
	int nestID, int tokCnt) {

	if (localTDec == null) {
	// nothing to worry about.
	    return;
	} // end if.

	String constr = localTDec.checkConditionals(myTypes, tokCnt);
	if (constr == null) {
	// nothing to do.
	    return;
	} // end if.

	// Else, load a dynamic constraint.
	int pos = constr.indexOf("FORCE_TOKEN");
	if (pos > -1) {
	// then this is a "force" constraint (as opposed to a "no-op");
	// we need to strip off the "FORCE_TOKEN" keyword for the
	// call to 'loadConstraint'.
	    constr = constr.substring(pos+11, constr.length());
	} // end if.

	// Now make the call to add the constraint.
	localTDec.loadConstraint(new StringTokenizer(constr), DYNAMIC);

	// Done.
	return;

    } // end method checkCondConstr.

    // ----

    /* ==========================================================
     *
     * method determineType():
     *
     * When we make this call, we assume that ALL tokens in this token
     * list object have been fully expanded and have result types
     * associated with them.  Using those types ('myTypes'), we first
     * verify that _all_ token constraints have been met, and then, if
     * that is in fact the case, we use the "types" and "op tables"
     * input files to figure out what the result type of this token
     * list node as a _whole_ is.  We then return this type, as well
     * as the final string ('str') that was ultimately generated by
     * expansion of this token list's tokens.
     *
     * -------- */

    public GResult determineType(String head, String str, int nestID,
	int syntax, String [] typesWhenEntered, TypeDecl entryLocalTD) {

	// Store the final string result.
	GResult res = new GResult();
	res.strVal = str;

	// Find out the appropriate 'head'.
	String fullHead;
	if (nestID != START_FRESH) {
	    fullHead = head + "_NESTED_" + syntax + "_" + nestID;
	} // end if.
	else {
	    fullHead = head;
	} // end else.

	// Find out what the final, overall result type is.
	if (localTDec == null) {
	// No type decl exists for this head; default 'type' is just the
	// head of the rule.
	    res.typeVal = head;
	} // end if.
	else {
	    if (!localTDec.verifyTypes(syntax, myTypes, gramSubTypes)) {
	    // then we're not semantically correct.
		res.strVal = "FAILED_GENERATION";
		res.typeVal = head;  // to avoid 'null'; not required.
	    } // end if.
	    else {
	    // we ARE semantically (type) correct; find out what the
	    // result type should be.
	        res.typeVal = localTDec.determineType(syntax, myTypes,
			gramOpTables);
		if (res.typeVal.equals("NONE")) {
		// couldn't figure it out from input files;
		// default 'type' is just the head of the rule.
		    res.typeVal = head;
		} // end inner if.
		// else, we figured out the result type and stored it.
	    } // end inner else.
	} // end outer else.

//	printTypeCheck(head, fullHead, res);	// Debugging.

	// Before returning, reset our "types" array so that we don't
	// interfere with other ('higher') generations of this
	// TokenListNode; ditto for 'localTDec'.
	restoreTypes(typesWhenEntered);
	restoreLocalTD(entryLocalTD);

	// Now, we're done; return the result.
	return res;

    } // end method determineType.

    // ----

    /* ==========================================================
     *
     * method lookup():
     *
     * Takes a String corresponding to some field from the
     * configuration file and returns the matching value,
     * as found in the configMap built from the configuration
     * file.  If no matching map entry is found, then a message
     * is printed and "0" is returned.
     *
     * -------- */

    public double lookup(String recurID) {

	Double dbl = (Double)configMap.get(recurID);
	if (dbl == null) {
	    System.out.println("\n-=- ERROR: Couldn't find a value for " +
		recurID + " in the config file. -=-\n");
	    return 0;
	} // end if.
	else {
	    return dbl.doubleValue();
	} // end else.

    } // end method lookup.

    // ----

    /* ==========================================================
     *
     * method restoreLocalTD():
     *
     * Set the localTDec to be a copy of the received TypeDecl; this
     * is done right before returning from a recursive 'generate' call,
     * to "restore" the localTDec to how it was before we entered (the
     * received TypeDecl is the localTDec as it was before we started
     * messing with it).
     *
     * -------- */

    public void restoreLocalTD(TypeDecl oldTD) {

	if (oldTD == null) {
	    localTDec = null;
	    return;
	} // end if.

	// Else, restore it.
	localTDec = new TypeDecl(oldTD);
	return;

    } // end method restoreLocalTD.

    // ----

    /* ==========================================================
     *
     * method restoreTypes():
     *
     * Set 'myTypes' to be a copy of the received array of types; this
     * is done right before returning from a recursive 'generate' call,
     * to "restore" the 'myTypes' array to how it was before we entered
     * (the received array is 'myTypes' as it was before we started
     * messing with it).
     *
     * -------- */

    public void restoreTypes(String [] oldTypes) {

	for (int i = 0; i < oldTypes.length; i++) {
	    myTypes[i] = oldTypes[i];
	} // end for.
	return;

    } // end method restoreTypes.

    // ----

    /* ==========================================================
     *
     * method findTypeDecl():
     *
     * Takes a head with syntax and a nested list id, and uses this
     * information to retrieve the corresponding global TypeDecl.
     * If the result is 'null', this must be handled by the calling
     * procedure.
     *
     * -------- */

    public TypeDecl findTypeDecl(String head, int syntax, int nestID) {

	TypeDecl tDec;
	if (nestID != START_FRESH) {
	    tDec = getTypeDecl(head + "_NESTED_" + syntax + "_" + nestID);
	} // end if.
	else {
	    tDec = getTypeDecl(head);
	} // end else.

	return tDec;

    } // end method findTypeDecl.

    /* ==========================================================
     *
     * [ Other Helper Functions ]:
     *
     * The remaining functions simply allow access to this
     * TokenListNode's data and statistics.
     *
     * -------- */

    public String toString() {

	StringBuffer str = new StringBuffer();
	Iterator itr = myTokens.iterator();
	while (itr.hasNext()) {
	    str.append(((TokenNode)itr.next()).toString());
	    str.append(" ");
        } // end while.

	return str.toString();

    } // end toString.

    // ----

    public Vector getList() {

	return myTokens;

    } // end method getList.

    // ----

    public TypeDecl getLocalTDec() {

	return localTDec;

    } // end method getList.

    // ----

    public int recurs() {

	return recurSymbols.length();

    } // end method recurs.

    // ----

    public String recurSymbols() {

	return recurSymbols;

    } // end method recurs.

    // ----

    public int countNestedLists() {

	int cnt = 0;
	TokenNode tmpTok;
	Iterator itr = myTokens.iterator();
	while (itr.hasNext()) {
	    tmpTok = (TokenNode)itr.next();
	    if (tmpTok.type() == DEF_LIST) {
	    // then count this one, plus any
	    // sublists that it might have.
		cnt++;
		cnt += (tmpTok.getList()).numNestedLists();
	    } // end if.
	} // end while.

	return cnt;

    } // end method countNestedLists.

    // ----

    // ===== Following method is for debugging only. =====

    public void printTypeCheck(String head, String fullHead,
	GResult res) {

//	if (head.equals("UnaryExpr")) {
	    System.out.print("Type check for " + fullHead + ": ");
	    System.out.print(res.typeVal + " w/ ");
	    for (int i = 0; i < myTypes.length; i++)
		System.out.print(myTypes[i] + " " );
	    System.out.println();
//	} // end if.
	return;

    } // end method printTypeCheck.

} // end class TokenListNode.
