import java.io.*;
import java.lang.Math;
import java.util.*;

// -----------------============== NOTE =============--------------------
// To make changes to "terminal" data (for grammars other than XQuery),
// go to the "Generator.java" file--that is the ONLY file that needs to
// be modified for those purposes; everything else is designed to run
// independently of the grammar being used.
// -----------------=================================--------------------

class TokenNode extends GrammarNode {

    // -------------------------------------------------------
    // 		   -=-=-=- TokenNode -=-=-=-
    // -------------------------------------------------------
    // A token node consists of one of four pieces of a rule's
    // definition: 1) an Id, 2) a String Literal, 3) a Character
    // Set, or 4) a nested DefListNode.  Each of these pieces can
    // have an occurence indicator attached to it; the piece
    // (content) and the occurrence indicator together constitute
    // a "token".  For example, in the following rule:
    //
    // fake ::= "{" def2+ [a-z] "}" (def3 def4)?
    //
    // there are FIVE tokens:
    //
    //		1) "{"		// String Literal, no occur indic.
    //		2) def2		// Id, "+" occur indic.
    //		3) [a-z]	// Character set, no occur indic.
    //		4) "{"		// String Literal, no occur indic.
    //		5) (def3 def4)?	// DefListNode, "?" occur indic.
    //
    // Naturally, when the nested DefListNode is processed, it is
    // recursively found to have two other tokens (def3 and def4),
    // both without occurrence indicators.
    //
    // -------------------------------------------------------
 
    // ************--==== Class Variables ====--************

    // Two kids: Content and Occurence Indicator.
    private TokenContentNode myContent;
    private TokenOccurNode myOccur;

    // Status variables.
    boolean hasOccur;
    boolean hasNestedOccur;

    // **********--==== End Class Variables ====--**********

    // ----

    /* ==========================================================
     *
     * Constructor 1.
     *
     * -------- */

    public TokenNode(TokenContentNode content) {

	myContent = content;
	hasOccur = false;
	hasNestedOccur = false;

    } // end constructor 1.

    // ----

    /* ==========================================================
     *
     * Constructor 2.
     *
     * -------- */

    public TokenNode(TokenContentNode content,
		TokenOccurNode occur) {

	myContent = content;
	myOccur = occur;
	hasOccur = true;
	hasNestedOccur = false;

    } // end constructor 2.

    // ----

    /* ==========================================================
     *
     * method unparse():
     *
     * Build and process this TokenNode structure based on inputs
     * and gathered stats, and then echo the structure to file.
     *
     * -------- */

    public void unparse(PrintWriter p, int indent) {

	myContent.unparse(p, indent);
	if (hasOccur) {
	    myOccur.unparse(p, indent);
	} // end if.

	// Look for nested occurence indicators.
	if (myContent.type() == DEF_LIST) {
	    DefListNode tmpList = myContent.getList();
	    if (tmpList.recurs() > 0) {
		hasNestedOccur = true;
	    } // end if.
	} // end if.

    } // end method unparse.

    // ----

    /* ==========================================================
     *
     * method charSelect():
     *
     * If the Content of this TokenNode is a character set, then
     * we want to pick one of the allowed values and return it
     * (as a string).
     *
     * -------- */

    public String charSelect() {

	if (type() != CHARSET) {
	    System.out.println("-=- ERROR: Tried to select" +
	    " a character from a non-CHARSET node. -=-\n");
	    return "BAD_CHAR!";
	} // end if.

	// Else, select a character.
	return myContent.charSelect();

    } // end method charSelect.

    // ----

    /* ==========================================================
     *
     * [ Other Helper Functions ]:
     *
     * The remaining functions simply allow access to this
     * TokenNode's data and statistics.
     *
     * -------- */

    public String toString() {

	String str = myContent.toString();
	if (hasOccur) {
	    str += myOccur.toString();
	} // end if.

	return str;

    } // end method toString.

    // ----

    public String idString() {

	return myContent.toString();

    } // end method toString.

    // ----

    public DefListNode getList() {

	if (type() != DEF_LIST) {
	    System.out.println("\n-=- ERROR: Tried to get" +
		" def list from a non-DefListNode. -=-\n");
	    return null;
	} // end if.
	else {
	    return myContent.getList();
	} // end else.

    } // end method getList.

    // ----

    public int type() {

	return myContent.type();

    } // end method type.

    // ----

    public boolean recurs() {

	return (hasOccur || hasNestedOccur);

    } // end method recurs.

    // ----

    public boolean recurs(int flag) {

	return (hasOccur);

    } // end method recurs.

    // ----

    public char recurType() {

	if (hasOccur) {
	    return myOccur.toChar();
	} // end if.
	else {
	    return NO_RECURRENCE;
	} // end else.

    } // end method recurs.

    // ----

    public int recurCnt() {

	int result = 0;
	if (type() == DEF_LIST) {
	    result = (myContent.getList()).recurs();
	} // end if.

	if (recurs(NON_NESTED_RECUR)) {
	    result++;
	} // end if.

	return result;

    } // end method recurCnt.

} // end class TokenNode
