import java.util.*;

/* *********************************************************************
 *
 * Grammar for declaring a TypeDecl is as follows:
 *
 * ==============================================
 * Grammar
 * ==============================================
 *
 * TypeDef         ::= "@" NonTerminal " ==> " (TypeSpec ";")* TypeSpec "."
 * TypeSpec        ::= Type | Dependency | Constraint | "NO_OP"
 * Type            ::= "TYPE " (TypeName | "OP_TABLE") SyntaxList?
 * Dependency      ::= "USE_TOKEN " TokenId SyntaxList?
 * Constraint      ::= "FORCE_TOKEN" ("NO_OP" | (TokenId ForceClause SyntaxList?))
 * ForceClause     ::= ("TO_TYPE" TypeList) | ("TO_TOKEN" TokenId)
 * TypeList        ::= TypeName (" " TypeName)*
 * SyntaxList      ::= "FOR_SYNTAX" SyntaxId (" " SyntaxId)*
 *
 * ******************************************************************* */

class TypeDecl {

    private String myHead;
    private Map myTypes;		// Integer -> String
    private Map myDependencies;		// Integer -> Integer
    private Map myTokConstraints;	// Integer -> Vector (of Constraint)
    private Map myTypeConstraints;	// Integer -> Vector (of Constraint)
    private Map myConditionals;		// Integer -> OpTable

    public static final int TYPE_CONSTRAINT = 0;
    public static final int TOK_CONSTRAINT = 1;

    protected static final boolean DYNAMIC = true;
    protected static final boolean NOT_DYNAMIC = false;

    // ----

    /* ==========================================================
     *
     * Constructor 1: Copy constructor.
     *
     * Create a (deep) copy of the received TypeDecl object.
     *
     * -------- */

    public TypeDecl(TypeDecl tDec) {

	myHead = tDec.getHead();
	myTypes = tDec.getMap(0);
	myDependencies = tDec.getMap(1);
	myConditionals = tDec.getMap(4);

	// For vectors, need to copy them.
	Map tmpMap = tDec.getMap(2);
	myTokConstraints = deepCopy(tmpMap);
	tmpMap = tDec.getMap(3);
	myTypeConstraints = deepCopy(tmpMap);

	// Done.
	return;

    } // end Constructor 1.

    // ----

    /* ==========================================================
     *
     * Constructor 2:
     *
     * Build a TypeDecl object out of the received string, which
     * should consist of all the necessary parts (if not, errors
     * will be generated and construction will be aborted).  The
     * received tDec should conform to the following syntax:
     *
     * TypeDef         ::= "@" NonTerminal " ==> " (TypeSpec ";")* TypeSpec "."
     *
     * -------- */

    public TypeDecl(String tDec) {

	// Load the head (NonTerminal)
	int headPos = tDec.indexOf("==>");
	if (headPos < 0) {
	    System.out.println("-=- ERROR: Invalid Type Declaration" +
		" Format (no head):\n" + tDec + " -=-");
	    myHead = "INVALID";
	    return;
	} // end if.

	// Find out where we end.
	int endPos = tDec.indexOf(".");

	// Now, initialize values, then go.
	try {
	    myHead = tDec.substring(0, headPos);
	    myHead = myHead.trim();
	    initMaps();
	    tDec = tDec.substring(headPos + 3, endPos);
	} catch (Exception e) {
	    System.out.println("-=- ERROR: Parse error in type file" +
		" for rule:\n\t" + tDec + "\n");
	    System.exit(0);
	} // end catch.

	// Now, load the type data.
	String delims = "\t ,";
	StringTokenizer tokenizer;
	String tmpStr;
	boolean okay = false;
	int currPos = 0;
	int tmpPos = tDec.indexOf(";", currPos);
	while (currPos >= 0) {
	// loop until entire decl is processed, TypeSpec by TypeSpec.

	    if (tmpPos < 0) {
	    // this is the last TypeSpec in the type decl; grab all of
	    // what's left.
		tmpStr = tDec.substring(currPos+1, tDec.length());
	    } // end if.
	    else {
	    // just grab the next TypeSpec in the type decl.
		tmpStr = tDec.substring(currPos+1, tmpPos);
	    } // end else.
	    tmpStr = tmpStr.trim();
	    tokenizer = new StringTokenizer(tmpStr, delims);
	    if (!tokenizer.hasMoreTokens()) {
		System.out.println("-=- ERROR: Invalid Type Declaration" +
		    " Format (missing type spec). -=-");
		myHead = "INVALID";
		initMaps();
		return;
	    } // end if.
	    // We have the next TypeSpec as a string; load it.
	    okay = doLoad(tokenizer);
	    if (!okay) {
	    // error somewhere; reset and get out.
		myHead = "INVALID";
		initMaps();
		return;
	    } // end if.

	    // else, continue.
	    currPos = tmpPos;
	    tmpPos = tDec.indexOf(";", currPos+1);

	} // end while.

	// By the time we get here, types, dependencies, and
	// constraints have all been loaded; we're done.
	return;

    } // end Constructor 2.

    // ----

    /* ==========================================================
     *
     * method initMaps():
     *
     * Initialize private data structures.
     *
     * -------- */

    private void initMaps() {

	myTypes = new HashMap();
	myDependencies = new HashMap();
	myTypeConstraints = new HashMap();
	myTokConstraints = new HashMap();
	myConditionals = new HashMap();

    } // end method initMaps.

    // ----

    /* ==========================================================
     *
     * method doLoad():
     *
     * Parse and load a TypeSpec from the received tokenizer.  The
     * tokenizer should conform to the following syntax:
     *
     * TypeSpec        ::= Type | Dependency | Constraint | "NO_OP"
     *
     * -------- */

    private boolean doLoad(StringTokenizer tokenizer) {

	String tmpTok = tokenizer.nextToken();
	if (tmpTok.equals("TYPE")) {
	// load a type.
	    return loadType(tokenizer);
	} // end if.
	else if (tmpTok.equals("USE_TOKEN")) {
	// load a dependency.
	    return loadDependency(tokenizer);
	} // end if.
	else if (tmpTok.equals("FORCE_TOKEN")) {
	// load a constraint.
	    return loadConstraint(tokenizer, NOT_DYNAMIC);
	} // end else if.
	else if (tmpTok.equals("NO_OP")) {
	// special indicator; don't do anything with this.
	    return true;
	} // end else if.
	else {
	// unknown token.
	    System.out.println("-=- ERROR: Invalid Type Declaration" +
		" Format (unknown spec: '" + tmpTok + "'). -=-");
	    return false;
	} // end else.

    } // end method doLoad.

    // ----

    /* ==========================================================
     *
     * method loadType():
     *
     * Parses and loads a type specification--i.e. a declaration
     * saying what the result type is for a specific syntax (or set
     * of syntaxes).  The received string tokenizer should conform to
     * the following syntax (with the "TYPE" keyword removed, since it
     * was already processed):
     *
     * Type            ::= "TYPE " (TypeName | "OP_TABLE") SyntaxList?
     *
     * The method returns 'true' if the type specification was
     * correctly parsed and loaded; 'false' otherwise.
     *
     * -------- */

    private boolean loadType(StringTokenizer strTok) {

	if (!strTok.hasMoreTokens()) {
	    System.out.println("-=- ERROR: Invalid Type Declaration" +
		    " Format (no types specified). -=-");
	    return false;
	} // end if.

	// Read type name.
	String tName = strTok.nextToken();

	// Read and load syntaxes corresponding to that type.
	if (!strTok.hasMoreTokens()) {
	// then ALL syntax choices are of this type; we indicate
	// this by using a syntax of "-1", which can't happen
	// otherwise.
	    myTypes.put(new Integer(-1), tName);
	    return true;
	} // end if.
	
	// Else: first, read in the 'FOR_SYNTAX' keyword (it's
	// there for clarification in the types file).
	String tmpStr = strTok.nextToken();
	if (!tmpStr.equals("FOR_SYNTAX") || !strTok.hasMoreTokens()){
	    System.out.println("-=- ERROR: Invalid Type Declaration" +
		    " Format (incomplete 'TYPE' spec). -=-");
	    return false;
	} // end if.

	// Finally, load the syntaxes listed for this type (SyntaxList).
	Integer tmpInt = null;
	while (strTok.hasMoreTokens()) {
	    try {
		tmpInt = Integer.valueOf(strTok.nextToken());
		myTypes.put(tmpInt, tName);
	    } // end try.
	    catch (NumberFormatException e) {
		System.out.println("-=- NOTE: Ignoring invalid syntax" +
		   " for type '" + tName + "'. -=-");
	    } // end catch.
	} // end while.

	return true;

    } // end method loadType

    // ----

    /* ==========================================================
     *
     * method loadDependency():
     *
     * Parses and loads a dependency specification--i.e. a declaration
     * saying that the result type for a specific syntax (or set
     * of syntaxes) depends on some token.  The received string
     * tokenizer should conform to the following syntax (with the
     * "USE_TOKEN" keyword removed, since it was already processed):
     *
     * Dependency      ::= "USE_TOKEN " TokenId SyntaxList?
     *
     * The method returns 'true' if the dependency specification
     * was correctly parsed and loaded; 'false' otherwise.
     *
     * -------- */

    private boolean loadDependency(StringTokenizer strTok) {

	if (!strTok.hasMoreTokens()) {
	    System.out.println("-=- ERROR: Invalid Dependency" +
		    " Format (no types specified). -=-");
	    return false;
	} // end if.

	// Read token on which we're dependent (TokenId).
	Integer tokId = null;
	try {
	    tokId = Integer.valueOf(strTok.nextToken());
	} // end try.
	catch (NumberFormatException e) {
	    System.out.println("-=- ERROR: Invalid Dependency" +
		   " Declaration (not a number). -=-");
	    return false;
	} // end catch.

	// Read and load syntaxes depending on that token.
	if (!strTok.hasMoreTokens()) {
	// then ALL syntax choices are of this type; we indicate
	// this by using a syntax of "-1", which can't happen
	// otherwise.
	    myDependencies.put(new Integer(-1), tokId);
	    return true;
	} // end if.
	
	// Else: first, read in the 'FOR_SYNTAX' keyword (it's
	// there for clarification in the types file).
	String tmpStr = strTok.nextToken();
	if (!tmpStr.equals("FOR_SYNTAX") || !strTok.hasMoreTokens()){
	    System.out.println("-=- ERROR: Invalid Dependency" +
		    " Format (incomplete 'USE_TOKEN' spec). -=-");
	    return false;
	} // end if.

	// Finally, load the syntaxes dependent on tokId.
	Integer tmpInt = null;
	while (strTok.hasMoreTokens()) {
	    try {
		tmpInt = Integer.valueOf(strTok.nextToken());
		myDependencies.put(tmpInt, tokId);
	    } // end try.
	    catch (NumberFormatException e) {
		System.out.println("-=- NOTE: Ignoring invalid entry" +
		   " for dependency on '" + tokId + "'. -=-");
	    } // end catch.
	} // end while.

	return true;

    } // end method loadDependency.

    // ----

    /* ==========================================================
     *
     * method loadConstraint():
     *
     * Parses and loads a constraint specification--i.e. a declaration
     * saying that for a specific syntax (or set of syntaxes), some
     * token must have a specific type in order to be semantically
     * correct.  The received string tokenizer should conform to the
     * following syntax (with the "FORCE_TOKEN" keyword removed, since
     * it was already processed):
     *
     * Constraint      ::= "FORCE_TOKEN" ("NO_OP" | (TokenId ForceClause SyntaxList?))
     * ForceClause     ::= ("TO_TYPE" TypeList) | ("TO_TOKEN" TokenId)
     *
     * The method returns 'true' if the type specification was
     * correctly parsed and loaded; 'false' otherwise.
     *
     * NOTE: Method is public so it can be called for the creation
     * of dynamic constraints (see 'TokenListNode.java').
     *
     * -------- */

    public boolean loadConstraint(StringTokenizer strTok,
		boolean isDynamic) {

	if (!strTok.hasMoreTokens()) {
	    System.out.println("-=- ERROR: Invalid Constraint " +
			"Declaration Format (no constraints " +
			"specified). -=-");
	    return false;
	} // end if.

	String tmpStr = strTok.nextToken();
	if (tmpStr.equals("NO_OP")) {
	// special indicator (should only be used with dynamic
	// constraints); don't do anything.
	    return true;
	} // end if.

	// Else, read in the number of the token that is being constrained.
	Integer constrained = null;
	try {
	    constrained = Integer.valueOf(tmpStr);
	} // end try.
	catch (NumberFormatException e) {
	    System.out.println("-=- ERROR: Invalid Constraint " +
			"Declaration Format (not an integer). -=-");
	    return false;
	} // end catch.

	// Read and load types to which that token is constrained.
	// Start by reading in the 'TO_...' keyword (needed to
	// indicate what kind of constraint we have).
	if (!strTok.hasMoreTokens()) {
	    System.out.println("-=- ERROR: Invalid Constraint " +
			"Declaration Format (incomplete). -=-");
	    return false;
	} // end if.
	tmpStr = strTok.nextToken();
	if ((!tmpStr.equals("TO_TYPE") && !tmpStr.equals("TO_TOKEN")) 
		|| !strTok.hasMoreTokens()) {
	    System.out.println("-=- ERROR: Invalid Constraint " +
		    "Format (incomplete). -=-");
	    return false;
	} // end if.

	// Load the constraints.  If this is a type constraint,
	// load list of types (strings) into myTypeConstraints; if it's
	// a token constraint, load a SINGLE token id (Integer) into
	// myTokConstraints.  Note that the "SyntaxList" part of the
	// declaration is handled within the following method calls.

	if (tmpStr.equals("TO_TYPE")) {
	// "TO_TYPE" ==> type constraints.
	    return loadTypeConstraint(strTok, constrained, isDynamic);
	} // end if.
	else {
	// "TO_TOKEN" ==> token constraints (there can only be ONE
	// constraint listed).
	    return loadTokConstraint(strTok, constrained, isDynamic);
	} // end else.

    } // end method loadConstraint.

    // ----

    /* ==========================================================
     *
     * method loadTypeConstraint():
     *
     * Parses and loads a type constraint specification--i.e. a
     * declaration saying that for a specific syntax (or set of
     * syntaxes), some token must have an explicitly stated type
     * in order to be semantically correct.  The received string
     * tokenizer should conform to the following syntax (with the
     * "TO_TYPE" keyword removed, since it was already processed):
     * 
     * TypeConstraint  ::= "TO_TYPE" TypeList SyntaxList?
     *
     * The method returns 'true' if the type specification was
     * correctly parsed and loaded; 'false' otherwise.
     *
     * -------- */

    private boolean loadTypeConstraint(StringTokenizer strTok,
		Integer constrained, boolean isDynamic) {

	// Load the list of types.
	Vector typeVec = new Vector();
	String tmpStr;
	while (strTok.hasMoreTokens()) {
	    tmpStr = strTok.nextToken();
	    if (tmpStr.equals("FOR_SYNTAX")) {
	    // then we're done with our list of types; move on.
		if (typeVec.size() == 0) {
		    System.out.println("-=- ERROR: No types " +
			"specified for type constraint! " + getHead() +" -=-");
		    return false;
		} // end if.
		return loadSyntaxList(strTok, TYPE_CONSTRAINT, typeVec, true);
	    } // end if.
	    typeVec.add(new Constraint(constrained, tmpStr, isDynamic));
	} // end while.

	return loadSyntaxList(strTok, TYPE_CONSTRAINT, typeVec, false);

    } // end method loadTypeConstraint.

    // ----

    /* ==========================================================
     *
     * method loadTokConstraint():
     *
     * Parses and loads a token constraint specification--i.e. a
     * declaration saying that for a specific syntax (or set of
     * syntaxes), some token must have the SAME type as another
     * (given) token in order to be semantically correct.  The
     * received string tokenizer should conform to the following
     * syntax (with the "TO_TOKEN" keyword removed, since it was
     * already processed):
     * 
     * TokenConstraint  ::= "TO_TOKEN" TokenId SyntaxList?
     *
     * The method returns 'true' if the type specification was
     * correctly parsed and loaded; 'false' otherwise.
     *
     * -------- */

    private boolean loadTokConstraint(StringTokenizer strTok,
		Integer constrained, boolean isDynamic) {

	Vector tokVec = new Vector();

	// First, get the target token.
	Integer tmpInt = null;
	try {
	    tmpInt = Integer.valueOf(strTok.nextToken());
	    tokVec.add(new Constraint(constrained, tmpInt, isDynamic));
	} // end try.
	catch (NumberFormatException e) {
	    System.out.println("-=- ERROR: Invalid token constraint" +
		    " on token '" + constrained + "'. -=-");
	} // end catch.
	if (strTok.hasMoreTokens()) {
	    if (strTok.nextToken().equals("FOR_SYNTAX")) {
	    // we're ready to load syntax list.
		return loadSyntaxList(strTok, 1, tokVec, true);
	    } // end if.
	    else {
		System.out.println("-=- WARNING: Extra token constraints" +
			" on token '" + constrained + "' were ignored. -=-");
	    } // end inner else.
	} // end if.

	return loadSyntaxList(strTok, TOK_CONSTRAINT, tokVec, false);

    } // end method loadTokConstraint.

    // ----

    /* ==========================================================
     *
     * method loadSyntaxList():
     *
     * Parses a list of syntax numbers and stores each syntax in
     * a constraint map with value equal to the received Vector
     * of Constraint objects.  If the list is empty, then a syntax
     * of '-1' is used to indicate that the received vector should
     * be mapped to ALL syntaxes (i.e. all syntaxes must abide by
     * the constraints contained in 'vec').
     *
     * -------- */

    boolean loadSyntaxList(StringTokenizer strTok, int constrType,
		 Vector vec, boolean hasSyntaxes) {

	if (!hasSyntaxes) {
	// then constraint applies to ALL syntaxes; use '-1'.
	    addConstraints(constrType, new Integer(-1), vec);
	    return true;
	} // end if.

	// Else, proceed to load syntaxes.
	Integer tmpInt = null;
	while (strTok.hasMoreTokens()) {
	    try {
		tmpInt = Integer.valueOf(strTok.nextToken());
		addConstraints(constrType, tmpInt, vec);
	    } // end try.
	    catch (NumberFormatException e) {
		System.out.println("-=- NOTE: Ignoring invalid entry" +
		   " for token constraint'. -=-");
	    } // end catch.
	} // end while.

	return true;

    } // end method loadSyntaxList.

    // ----

    /* ==========================================================
     *
     * method addConstraints():
     *
     * Adds an entry to either myTypeConstraints or myTokConstraints
     * (depending on the received constrType) that maps the received
     * syntax to the received vector of constraints.
     *
     * -------- */

    public void addConstraints(int constrType, Integer syntax,
		Vector constraints) {

	// First, remove useless constraints.
	Vector newConstraints = removeUselessFrom(constraints, syntax);

	if (constrType == TYPE_CONSTRAINT) {
	    Vector allConstraints =
		(Vector)myTypeConstraints.get(syntax);
	    if (allConstraints != null) {
	    // then an entry already exists; add to what's already there
	    // (put final result in 'allConstraints').
		doUnion(allConstraints, newConstraints, constrType);
		myTypeConstraints.put(syntax, allConstraints);
	    } // end inner if.
	    else {
		myTypeConstraints.put(syntax, newConstraints);
	    } // end inner else.
	} // end if.
	else { // token constraint.
	    // See if there are already constraints for this syntax.
	    Vector allConstraints =
		(Vector)myTokConstraints.get(syntax);
	    if (allConstraints != null) {
	    // then an entry already exists; add to what's already there
	    // (put final result in 'allConstraints').
		doUnion(allConstraints, newConstraints, constrType);
		myTokConstraints.put(syntax, allConstraints);
	    } // end inner if.
	    else {
		myTokConstraints.put(syntax, newConstraints);
	    } // end inner else.
	} // end else.

	// Done.
	return;

    } // end method addConstraint.

    // ----

    /* ==========================================================
     *
     * method removeUselessFrom():
     *
     * Removes any constraints from the received vector that have
     * victims matching any other constraints in this TypeDecl (for
     * the same syntax); we ignore such constraints because it
     * doesn't make sense to try to force the same victim token to
     * two different constraints.
     *
     * -------- */

    public Vector removeUselessFrom(Vector constr, Integer syntax) {

	Vector res = new Vector();
	Iterator itr = constr.iterator();
	while (itr.hasNext()) {
	    Constraint c = (Constraint)itr.next();
	    if (!(constraintFor(c.victim(), syntax, TYPE_CONSTRAINT)) &&
		!(constraintFor(c.victim(), syntax, TOK_CONSTRAINT))) {
	    // then we keep the constraint around.
		res.add(c);
	    } // end if.
	    // else, some constraint already exists for the current
	    // syntax on the victim of this constraint; so ignore this
	    // constraint.
	} // end while.

	// Return the result.
	return res;

    } // end method constraintExists.

    // ----

    /* ==========================================================
     *
     * method constraintFor():
     *
     * Returns true if a constraint of the received kind exists
     * on the received token for the received syntax; otherwise,
     * returns false.
     *
     * -------- */

    public boolean constraintFor(int tok, Integer syntax, int kind) {

	Vector vec;
	if (kind == TYPE_CONSTRAINT) {
	    vec = (Vector)myTypeConstraints.get(syntax);
	    if (vec == null) {
	    // none for the specific syntax; check for general.
		vec = (Vector)myTypeConstraints.get(new Integer(-1));
	    } // end if.
	} // end if.
	else {
	    vec = (Vector)myTokConstraints.get(syntax);
	    if (vec == null) {
	    // none for the specific syntax; check for general.
		vec = (Vector)myTokConstraints.get(new Integer(-1));
	    } // end if.
	} // end else.

	if (vec == null) {
	// then we can't possibly have any duplicate constraints.
	    return false;
	} // end if.

	// Else...
	Iterator itr = vec.iterator();
	while (itr.hasNext()) {
	    if (((Constraint)itr.next()).victim() == tok) {
	    // then a constraint already exists.
		return true;
	    } // end if.
	} // end while.

	// If we get here, no constraint already exists.
	return false;

    } // end method constraintFor.

    // ----

    /* ==========================================================
     *
     * method doUnion():
     *
     * Takes two vectors (of Constraint objects) and forms their
     * union, storing the result in the _first_ vector.  This
     * means that ORDER DOES matter: v1 will be changed and will
     * hold the final result, while v2 stays as it is.
     *
     * NOTE: we don't worry about duplicate constraints in this
     * method because if any exist, they should have been removed
     * from v2 by now (via a call to "removeUselessFrom()" in
     * the "addConstraints()" method.
     *
     * -------- */

    public void doUnion(Vector v1, Vector v2, int kind) {

	// Iterate through v2...
	Iterator itr2 = v2.iterator();
	while (itr2.hasNext()) {
	    v1.add(itr2.next());
	} // end outer while.

	// Done.
	return;

    } // end method doUnion.

    // ----

    /* ==========================================================
     *
     * method loadCondConstr():
     *
     * Receives a conditional constraint (in the form of an OpTable)
     * and adds it to this TypeDecl's map of such constraints.
     *
     * -------- */

    public void loadCondConstr(OpTable condTable, Integer baseTok) {

	// Check validity of inputs.
	if ((condTable == null) || (baseTok == null)) {
	// shouldn't happen.
	    System.out.println("-=- ERROR: Tried to add an invalid" +
		" conditional constraint to TypeDecl for '" +
		getHead() + "'. -=-");
	    return;
	} // end if.

	// Put it in our list.
	myConditionals.put(baseTok, condTable);

	// Done.
	return;

    } // end method loadCondConstr.

    // ----

    /* ==========================================================
     *
     * method checkConditionals():
     *
     * Runs through the conditional constraint table corresponding to
     * the received tokCnt and returns the constraint decl for the
     * FIRST row that matches the received 'types' array.  If no
     * row matches, then null will be returned.
     *
     * -------- */

    public String checkConditionals(String [] types, int tokCnt) {

	OpTable oTab = (OpTable)(myConditionals.get(new Integer(tokCnt)));
	if (oTab == null) {
	// no conditionals relevant to this constraint.
	    return null;
	} // end if.

	// Note that "determineType" in this case is actually "determine
	// the appropriate constraint".
	return oTab.determineType(types);

    } // end method CheckConditionals.

    // ----

    /* ==========================================================
     *
     * method getValidChoices():
     *
     * Using the received 'constraints' and 'subtypes' vectors, this
     * method looks at the result type specified for each syntax of
     * myHead and returns a list of the syntax choices that can
     * POTENTIALLY fulfill the received constraints (either directly
     * or by subtypes).
     *
     * -------- */

    public Vector getValidChoices(Vector constraints,
	Vector subtypes) {

	Vector res = new Vector();
	String typeStr = null;
	Integer syntax = null;

	// Get all subtypes of the types in constraints, so that we can
	// do a full check.
	Vector fullConstraints =
	     Subtype.getAllSubtypesIn(constraints, subtypes);

	// See which types are explicitly valid.
	Set typeSet = myTypes.keySet();
	Iterator itr = typeSet.iterator();
	while (itr.hasNext()) {
	    syntax = (Integer)itr.next();
	    typeStr = (String)myTypes.get(syntax);
	    if (fullConstraints.contains(typeStr) &&
	 	!res.contains(syntax)) {
	    // then this syntax conforms to the constraints;
	    // put it in our running list.
		res.add(syntax);
	    } // end if.
	    else if (typeForSyntax(syntax.intValue()).equals("OP_TABLE")) {
	    // Add any choices that depend on an Operation Table,
	    // since the result in the ops table MIGHT be valid.
		if (!res.contains(syntax)) {
		// don't add it twice.
		    res.add(syntax);
		} // end inner if.
	    } // end else if.
	} // end while.

	// Now, see which types might be valid indirectly (because
	// they depend on some other token, and that token MIGHT
	// be valid).
	Set depSet = myDependencies.keySet();
	itr = depSet.iterator();
	while (itr.hasNext()) {
	    syntax = (Integer)itr.next();
	    if (!res.contains(syntax)) {
	    // don't add it twice.
		res.add(syntax);
	    } // end if.
	} // end while.

	// Done.
	return res;

    } // end method getValidChoices.

    // ----

    /* ==========================================================
     *
     * method getConstraints():
     *
     * Returns a vector of constraints that apply to the received
     * tokenId for the received syntax.  All constraints are
     * returned as strings, so if tokenId is constrained to the
     * type of some other token, the TYPE of that other token will
     * be included in the result (not the token itself).
     * 
     * NOTE: we look for token constraints first, and if we find
     * them, we return--so in the event that token constraints AND
     * type constraints apply to tokenId (which doesn't make any
     * sense), we will always use the token constraints.
     *
     * -------- */

    public Vector getConstraints(int tokenId, String [] typeList,
	int syntax) {

	Vector res = new Vector();
	String tmpStr = null;
	Integer tmpInt = null;
	Constraint tmpConstraint = null;
	Vector tmpVec;

	// First, check for Token constraints.
	res = checkForTokConstraints(tokenId, typeList, syntax);
	if (res != null) {
	// then we found relevant token constraints; we're done.
	    return res;
	} // end if.

	// If we get here, we found no token constraints; check
	// for type constraints.
	res = new Vector();
	tmpVec = (Vector)(myTypeConstraints.get(new Integer(syntax)));
	if (tmpVec == null) {
	// see if we have a general constraint.
	    tmpVec = (Vector)(myTypeConstraints.get(new Integer(-1)));
	} // end if.
	if (tmpVec == null) {
	// no type constraints, either.  Return null.
	    return null;
	} // end if.

	// Else, load the type constraints for this tokenId.
	Iterator itr = tmpVec.iterator();
	while (itr.hasNext()) {
	    tmpConstraint = (Constraint)itr.next();
	    if (tmpConstraint.victim() == tokenId) {
	    // then this constraint applies to the token in question;
	    // put it in our running list (as long as it's NOT a force
	    // to "NONE"--if it is, we ignore it).
		if (!(tmpConstraint.asType().equals("NONE")) &&
		  !res.contains(tmpConstraint.asType())) {
		    res.add(tmpConstraint.asType());
		} // end inner if.
	    } // end if.
	} // end while.

	if (res.size() == 0) {
	// then we found no constraints on the received token within
	// the received syntax.
	    return null;
	} // end if.

	// Else, return the found type constraints.
	return res;		

    } // end method getConstraints.

    // ----

    /* ==========================================================
     *
     * method checkForTokConstraints():
     *
     * Returns a vector of constraints that apply to the received
     * tokenId for the received syntax.  All constraints are
     * returned as strings, so if tokenId is constrained to the
     * type of some other token (which is why we're here), the TYPE
     * of that other token will be included in the result (not the
     * token itself).
     *
     * -------- */

    public Vector checkForTokConstraints(int tokenId, String [] typeList,
	int syntax) {

	Vector res = new Vector();
	Constraint tmpConstraint = null;
	Vector tmpVec = (Vector)(myTokConstraints.get(new Integer(syntax)));
	if (tmpVec == null) {
	// see if we have a general constraint.
	    tmpVec = (Vector)(myTokConstraints.get(new Integer(-1)));
	} // end if.
	if (tmpVec != null) {
	// then we have token constraints on this syntax.
	    Iterator itr = tmpVec.iterator();
	    while (itr.hasNext()) {
		tmpConstraint = (Constraint)itr.next();
		if (tmpConstraint.victim() == tokenId) {
		// then this constraint applies to the token in question;
		// since we can only have ONE token constraint, store this
		// result and return (i.e. if there are multiple token
	 	// constraints on this token, we ONLY CONSIDER THE FIRST
		// ONE).
		    if (tmpConstraint.asToken() >= typeList.length) {
		    // bad constraint; just ignore it here (we'll print
		    // a message later, when we verify).
			  continue;
		    } // end inner-most if.
		    if (!(typeList[tmpConstraint.asToken()]).equals("NONE")) {
		    // never add "NONE" as a valid constraint...
			res.add(typeList[tmpConstraint.asToken()]);
		    } // end if.
		    return res;
		} // end inner if.
	    } // end while.
	} // end if.

	// If we get here, no token constraints; return indicator.
	return null;

    } // end method checkForTokConstraints.

    // ----

    /* ==========================================================
     *
     * method getTokConstraints():
     *
     * Return a vector of the token constraints that apply to the
     * received tokenId within the received syntax.
     *
     * -------- */

    public Vector getTokConstraints(int tokenId, int syntax) {

	Constraint tmpConstraint = null;
	Vector res = (Vector)(myTokConstraints.get(new Integer(syntax)));
	if (res == null) {
	// see if we have a general constraint.
	    res = (Vector)(myTokConstraints.get(new Integer(-1)));
	} // end if.

	return res;

    } // end method getTokConstraints.

    // ----

    /* ==========================================================
     *
     * method determineType():
     *
     * Look through this TypeDecl's constraints and find out the
     * final type for the received syntax.  If the type is given in
     * an OpTable, then look it up; if it's given as a dependency,
     * use the type of the token on which it's dependent; if it's
     * an explicit type name, then just return that.  If all of
     * this fails, return "NONE" as the type.
     *
     * -------- */

    public String determineType(int syntax, String [] tokTypes,
		Vector gramOpTables) {

	String currType = (String)myTypes.get(new Integer(syntax));
	if (currType == null) {
	// no type specified for this syntax; see if there's a
	// default one.
	    currType = (String)myTypes.get(new Integer(-1));
	} // end if.

	// See what we have...
	if (currType == null) {
	// Either we have a dependency, or we have no type.
	    // Check for dependency.
	    Integer tmpInt = (Integer)(myDependencies.get(new
					Integer(syntax)));
	    if (tmpInt != null) {
	    // then we found a dependency; use it to find type.
		if (tmpInt.intValue() >= tokTypes.length) {
		// bad force.
		    System.out.println("-=- ERROR: Invalid type" +
			" dependency:\n-=-\ttoken " + tmpInt.intValue() +
			" does not exist for '" + myHead +
			"' [ignored]. -=-");
		    return "NONE";
		} // end if.
		return tokTypes[tmpInt.intValue()];
	    } // end inner if.
	    else {
	    // see if our dependency is for ALL syntaxes.
		tmpInt = (Integer)(myDependencies.get(new Integer(-1)));
		if (tmpInt != null) {
		// then we found a dependency; use it.
		    if (tmpInt.intValue() >= tokTypes.length) {
		    // bad force.
			System.out.println("-=- ERROR: Invalid type" +
			  " dependency:\n-=-\ttoken " + tmpInt.intValue() +
			  " does not exist for '" + myHead +
			  "' [ignored]. -=-");
			return "NONE";
		    } // end inner if.
		    return tokTypes[tmpInt.intValue()];
		} // end if.
		// Else, we have no type.
		return "NONE";
	    } // end inner else.
	} // end outer if.
	else if (currType.equals("OP_TABLE")) {
	    OpTable opT = GrammarNode.getOpTable(myHead, gramOpTables);
	    if (opT == null) {
	    // no op table exists for this rule; nothing we can do.
		return "NONE";
	    } // end if.
	    return opT.determineType(tokTypes);
	} // end else if.
	else {
	// we found an explicit type; return it.
	    return currType;
	} // end else.

    } // end method determineType.

    // ----

    /* ==========================================================
     *
     * method verifyTypes():
     *
     * Look at the received token types for the received syntax and,
     * by looking through ALL relevant constraints in this TypeDecl, make
     * sure that the received tokTypes are in fact semantically valid.
     *
     * -------- */

    public boolean verifyTypes(int syntax, String [] tokTypes,
	Vector subtypes) {

	// First, verify Token constraints.
	boolean res = verifyTokConstraints(syntax, tokTypes, subtypes);
	if (!res) {
	// then we've already failed the type constraints; no need to go on.
	    return false;
	} // end if.

	// Now, verify Type constraints.
	Constraint tmpConstraint = null;
	Vector tmpVec = (Vector)(myTypeConstraints.get(new Integer(syntax)));
	if (tmpVec == null) {
	// see if we have a general constraint.
	    tmpVec = (Vector)(myTypeConstraints.get(new Integer(-1)));
	} // end if.
	if (tmpVec != null) {
	// then we have type constraints; check them.
	    String toCheck;
	    boolean pertains;
	    boolean isOkay;
	    for (int i = 0; i < tokTypes.length; i++) {
		toCheck = tokTypes[i];
		if (toCheck.equals("NONE")) {
		// if current token is of type "NONE", then we assume it is
		// valid; go to the next one.  NOTE: This can only happen
		// with optional tokens (i.e. tokens with occurence indicators
		// that allow them to _not_ be generated); if we count this
		// as a type and check it, then anytime the token is NOT
		// generated, we will fail the check--which is wrong.
		    continue;
		} // end if.
		pertains = false;
		isOkay = false;
		Iterator itr = tmpVec.iterator();
		while (itr.hasNext()) {
		    tmpConstraint = (Constraint)itr.next();
		    if (tmpConstraint.victim() == i) {
		    // then this constraint applies to the token in question;
		    // verify it--NOTE that if there are multiple type
		    // constraints on this token, we ONLY VERIFY the FIRST
		    // ONE.
			pertains = true;
			Subtype sub =
			   Subtype.getDeclFor(tmpConstraint.asType(),
				subtypes);
			if ((sub == null) &&
			  (tmpConstraint.asType()).equals(toCheck)) {
		        // we're okay for this token; continue.
			    isOkay = true;
			    break;  // from while.
			} // end if.
			else if ((sub != null) && sub.matches(toCheck)) {
		        // we're okay for this token; continue.
			    isOkay = true;
			    break;  // from while.
			} // end inner-most if.
		    } // end inner if.
		} // end while.
		if (pertains && !isOkay) {
		// we failed verification.
		    return false;
		} // end if.
	    } // end for.
	} // end outer if.

	// If we get here, either we have no constraints to worry about,
 	// or we have satisfied all of the constraints that we have.
	return true;

    } // end method verifyTypes.

    // ----

    /* ==========================================================
     *
     * method verifyTokConstraints():
     *
     * Look at the received token types for the received syntax and,
     * by looking through ALL relevant token constraints in this TypeDecl,
     * make sure that the received tokTypes are in fact semantically valid.
     * Return false if they are not.
     *
     * -------- */

    public boolean verifyTokConstraints(int syntax, String [] tokTypes,
	Vector subtypes) {

	Constraint tmpConstraint = null;
	Vector tmpVec = (Vector)(myTokConstraints.get(new Integer(syntax)));
	if (tmpVec == null) {
	// see if we have a general constraint.
	    tmpVec = (Vector)(myTokConstraints.get(new Integer(-1)));
	} // end if.
	if (tmpVec != null) {
	// then we have token constraints on this syntax.
	    String toCheck;
	    for (int i = 0; i < tokTypes.length; i++) {
		toCheck = tokTypes[i];
		if (toCheck.equals("NONE")) {
		// if current token is of type "NONE", then we assume it is
		// valid; go to the next one.  NOTE: This can only happen
		// with optional tokens (i.e. tokens with occurence indicators
		// that allow them to _not_ be generated); if we count this
		// as a type and check it, then anytime the token is NOT
		// generated, we will fail the check--which is wrong.
		    continue;
		} // end if.
		Iterator itr = tmpVec.iterator();
		while (itr.hasNext()) {
		    tmpConstraint = (Constraint)itr.next();
		    if (tmpConstraint.asToken() >= tokTypes.length) {
		    // constraint doesn't apply; print a message and
		    // ignore it.
			System.out.println("-=- ERROR: Invalid constraint:\n" +
			   "-=-\t token " + tmpConstraint.asToken() +
			   " does not exist for '" + myHead +
			   "' [ignored]. -=-");
			break;
		    } // end if.
		    if (tmpConstraint.victim() == i) {
		    // then this constraint applies to current token.  NOTE
		    // that if there are multiple token constraints on this
		    // token, we ONLY VERIFY the FIRST ONE.
			Subtype sub =
			   Subtype.getDeclFor(tokTypes[tmpConstraint.asToken()],
				subtypes);
			if ((sub == null) &&
			  !(tokTypes[tmpConstraint.asToken()]).equals(toCheck)) {
			// types don't check.
			    return false;
			} // end if.
			else if ((sub != null) && !sub.matches(toCheck)) {
			// types don't check.
			    return false;
			} // end inner if.
			break;	// from while.
		    } // end if.
		} // end while.
	    } // end for.
	} // end outer if.

	// If we get here, we good.
	return true;

    } // end method verifyTokConstraints.

    // ----

    /* ==========================================================
     *
     * method hasDepFor():
     *
     * Returns true if this TypeDecl has a dependency that applies
     * to the received syntax.
     *
     * -------- */

    public boolean hasDepFor(int syntax) {

	// Check the dependencies.
	if (myDependencies.containsKey(new Integer(syntax))) {
	// then we have a dependency for the received syntax;
	// thus, we are currently undecided.
	    return true;
	} // end if.

	if (myDependencies.containsKey(new Integer(-1))) {
	// then we have a dependency for ALL syntaxes;
	// thus, we are currently undecided.
	    return true;
	} // end if.

	if ((typeForSyntax(syntax) != null) && 
	   (typeForSyntax(syntax).equals("OP_TABLE"))) {
	// treat 'op table' as a dependency, since the type
	// "depends" on the op table.
	    return true;
	} // end if.

	// Else, we have no dependency..
	return false;

    } // end method hasDepFor.

    // ----

    /* ==========================================================
     *
     * method tokDependency():
     *
     * If the received syntax depends on a specific token for its
     * type, then return that token Id (as an Integer).  Otherwise,
     * return null.
     *
     * -------- */

    public Integer tokDependency(int syntax) {

	// Check for dependency.
	Integer tmpInt = (Integer)(myDependencies.get(new
				Integer(syntax)));
	if (tmpInt == null) {
	// nothing for the specific syntax received; check for
	// default.
	    tmpInt = (Integer)(myDependencies.get(new
				Integer(-1)));
	} // end if.

	if (tmpInt != null) {
	// then we found a dependency; return the token on which
	// the type depends.
	    return tmpInt;	
	} // end if.

	// Else, nothing.
	return null;

    } // end method tokDependency.

    // ----

    /* ==========================================================
     *
     * method typeForSyntax():
     *
     * Returns the explicit type for the received syntax.  By 
     * "explicit", we mean that the type is declared with the
     * TYPE keyword in the types file (thus, "OP_TABLE" can be
     * returned here, but not dependencies).  If no such type
     * is found, then null is returned.
     *
     * -------- */

    public String typeForSyntax(int syntax) {

	String result = (String)myTypes.get(new Integer(syntax));
	if (result == null) {
	// no type specified for this syntax; see if there's a
	// default one.
	    result = (String)myTypes.get(new Integer(-1));
	} // end if.
	
	return result;

    } // end method typeForSyntax.

    // ----

    /* ==========================================================
     *
     * method removeDynConstraints():
     *
     * Remove all dynamic constraints from this this type decl.
     * (NO LONGER USED, but kept it around anyways, just in case).
     *
     * -------- */

    public void removeDynConstraints() {

	// First, remove any dynamic token constraints.
	clearDynamic(myTokConstraints);

	// Then, remove any dynamic type constraints.
	clearDynamic(myTypeConstraints);

	// That's it.
	return;

    } // end method removeDynConstraints.

    // ----

    /* ==========================================================
     *
     * method clearDynamic():
     *
     * Clear out dynamic constraints from the received map.
     * (NO LONGER USED, but kept it around anyways, just in case).
     *
     * -------- */

    public void clearDynamic(Map aMap) {

	Constraint tmpCnst;
	Map.Entry tmpSet;
	Iterator itr = (aMap.entrySet()).iterator();
	while (itr.hasNext()) {
	    tmpSet = (Map.Entry)itr.next();
	    Vector tmpVec = (Vector)tmpSet.getValue();
	    Iterator itr2 = tmpVec.iterator();
	    int pos = 0;
	    while (itr2.hasNext()) {
		tmpCnst = (Constraint)itr2.next();
		if (tmpCnst.isDynamic()) {
		// then this is a dynamic constraint; remove it.
		    tmpVec.removeElementAt(pos);
		    // now we need start again, since we just
		    // changed the vector we're working with.
		    itr2 = tmpVec.iterator();
		    pos = -1;
		} // end if.
		pos++;
	    } // end inner while.
	    if (tmpVec.size() == 0) {
	    // then all constraints in this entry were dynamic; since
	    // we've removed them all, take out the entire entry, as well.
		aMap.remove(tmpSet.getKey());
		itr = (aMap.entrySet()).iterator();
	    } // end if.
	} // end while.

	// Done.
	return;

    } // end method removeDynConstraints.

    // ----

    public String toString() {

	return (myHead + " ::=\n\tTypes => " + myTypes +
		"\n\tDependencies => " + myDependencies + 
		"\n\tTokConstraints => " + myTokConstraints +
		"\n\tTypeConstraints => " + myTypeConstraints);

    } // end method toString.

    // ----

    public String getHead() {

	return myHead;

    } // end method getHead.

    // ----

    public Map getMap(int which) {

	switch(which) {

	    case 0: return myTypes;
	    case 1: return myDependencies;
	    case 2: return myTokConstraints;
	    case 3: return myTypeConstraints;
	    case 4: return myConditionals;
	    default:
		System.out.println("-=- ERROR: Request for non-existant" +
		    " map in TypeDecl. -=-");
		return null;

	} // end switch.

    } // end method getMap.

    // ----

    public Map deepCopy(Map srcMap){

	Map result = new HashMap();
	Iterator itr = (srcMap.entrySet()).iterator();
	Vector tmpVec;
	while (itr.hasNext()) {
	    Map.Entry entry = (Map.Entry)itr.next();
	    tmpVec = new Vector();
	    tmpVec.addAll((Vector)(entry.getValue()));
	    result.put((Integer)entry.getKey(), tmpVec);
	} // end while.

	return result;

    } // end method deepCopy.

} // end class TypeDecl.
